import logging
import os
import pickle
import random
import warnings
from functools import partial
from collections import defaultdict
from multiprocessing.pool import Pool

import numpy as np
import pyarrow.parquet as pq
import torch
from torch.utils.data import WeightedRandomSampler, Sampler, Dataset
from torch.utils.data.dataloader import DataLoader

from ptls.data_load.augmentations.all_time_shuffle import AllTimeShuffle
from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset
from ptls.data_load.padded_batch import PaddedBatch

logger = logging.getLogger(__name__)


def default_preprocess(data, conf, drop_unknown=True):
    random.seed(conf.params.seed)

    lag_days = conf.params.lag_days
    max_days = conf.params.max_days
    transaction_start_date = np.datetime64(conf.dataset.get('transaction_start_date', '1970-01-01T00:00'))

    for rec in data:
        target = rec['target']
        if target == -1 and drop_unknown:
            continue

        zero_take_prob = rec.get('zero_take_prob', 1.)

        if target == 0 and zero_take_prob < 1.0 and random.random() > zero_take_prob:
            continue

        feature_arrays = rec['feature_arrays']

        min_seq_len = rec.get('min_seq_len', 0)

        tranz_dates = rec['event_time']
        n_tranz = len(tranz_dates)
        if n_tranz < min_seq_len:
            continue

        rec['seq_len'] = n_tranz
        if n_tranz > 0:
            lagged_application_date = rec['application_date'] - np.timedelta64(lag_days, 'D')
            first_tranz_date = lagged_application_date - np.timedelta64(max_days, 'D')
            first_tranz_date = max(transaction_start_date, first_tranz_date)

            app_date_pos = np.searchsorted(tranz_dates, lagged_application_date)
            window_start_pos = np.searchsorted(tranz_dates, first_tranz_date)

            if (app_date_pos - window_start_pos) < min_seq_len:
                continue

            feature_arrays_tl = {
                k: arr[window_start_pos:app_date_pos]
                for k, arr in feature_arrays.items()
            }
            rec['feature_arrays'] = feature_arrays_tl
            rec['event_time'] = tranz_dates[window_start_pos:app_date_pos]

        yield rec


def read_spark_dataset_gen(data_files):
    import sparkpickle
    for path in data_files:
        with open(path, 'rb') as f:
            for row in sparkpickle.load_gen(f):
                yield dict(row)


def data_read_main(path, prepare_gen):
    import sparkpickle
    with open(path, 'rb') as f:
        rec_gen = (dict(e) for e in sparkpickle.load_gen(f))
        prepared_rec_gen = prepare_gen(rec_gen)
        return list(prepared_rec_gen)


def read_dataset_mthread(files, n_workers, prepare_gen):
    if n_workers > 1:
        with Pool(n_workers) as p:
            for chunk in p.imap_unordered(partial(data_read_main, prepare_gen=prepare_gen), files):
                for rec in chunk:
                    yield rec
    else:
        for rec in prepare_gen(read_spark_dataset_gen(files)):
            yield rec


def to_torch_compatible(a):
    if a.dtype == np.int8:
        return a.astype(np.int16)
    return a


def features2torch(seq):
    for rec in seq:
        rec['feature_arrays'] = {k: torch.from_numpy(to_torch_compatible(v)) for k, v in rec['feature_arrays'].items()}
        yield rec


def read_pyarrow_file(path, use_threads=True):
    p_table = pq.read_table(
        source=path,
        use_threads=use_threads,
    )

    col_indexes = [n for n in p_table.column_names]

    def get_records():
        for rb in p_table.to_batches():
            col_arrays = [rb.column(i) for i, _ in enumerate(col_indexes)]
            col_arrays = [a.to_numpy(zero_copy_only=False) for a in col_arrays]
            for row in zip(*col_arrays):
                # np.array(a) makes `a` writable for future usage
                # rec = {n: np.array(a) if isinstance(a, np.ndarray) else a for n, a in zip(col_indexes, row)}
                rec = {}
                for n, a in zip(col_indexes, row):

                    if n == 'trans_time':
                        date_batch = np.array(a).astype('datetime64[s]')
                        rec['local_day'] = (np.array(date_batch).astype('datetime64[D]') - np.array(date_batch))\
                                                                .astype('datetime64[M]').astype(np.int16) + 1
                        rec['local_month'] = date_batch.astype('datetime64[M]').astype(np.int16) / 12 + 1
                        rec['local_weekday'] = (date_batch.astype('datetime64[D]').astype(np.int16) + 3) / 7
                        rec['hour'] = ((date_batch - date_batch.astype('datetime64[D]')).astype(np.int32) // 3600 + 1)\
                                                                                      .astype(np.int16)
                    rec[n] = np.array(a) if isinstance(a, np.ndarray) else a
                yield rec

    return get_records()


def read_data_gen(path):
    ext = os.path.splitext(path)[1]
    if ext == '.p':
        with open(path, 'rb') as f:
            data = pickle.load(f)
            return iter(data)
    elif ext == '.parquet':
        return read_pyarrow_file(path, True)
    else:
        raise NotImplementedError(f'Unknown input file extension: "{ext}"')


class DropoutTrxDataset(Dataset):
    def __init__(self, dataset: Dataset, trx_dropout, seq_len, with_target=True):
        self.core_dataset = dataset
        self.trx_dropout = trx_dropout
        self.max_seq_len = seq_len
        self.style = dataset.style
        self.with_target = with_target

    def __len__(self):
        return len(self.core_dataset)

    def __iter__(self):
        for rec in iter(self.core_dataset):
            yield self._one_item(rec)

    def __getitem__(self, idx):
        item = self.core_dataset[idx]
        if type(item) is list:
            return [self._one_item(t) for t in item]
        else:
            return self._one_item(item)

    def _one_item(self, item):
        if self.with_target:
            x, y = item
        else:
            x = item

        seq_len = len(next(iter(x.values())))

        if self.trx_dropout > 0 and seq_len > 0:
            idx = np.random.choice(seq_len, size=int(seq_len * (1 - self.trx_dropout)+1), replace=False)
            idx = np.sort(idx)
        else:
            idx = np.arange(seq_len)

        idx = idx[-self.max_seq_len:]
        new_x = {k: v[idx] for k, v in x.items()}

        if self.with_target:
            return new_x, y
        else:
            return new_x


class LastKTrxDataset(Dataset):
    def __init__(self, dataset: Dataset, share):
        self.core_dataset = dataset
        self.share = share

    def __len__(self):
        return len(self.core_dataset)

    def __getitem__(self, idx):
        x, y = self.core_dataset[idx]

        seq_len = len(next(iter(x.values())))

        if self.share < 1.:
            start_idx = int(seq_len*(1-self.share))
            idx_to_take = list(range(seq_len))[start_idx:]
            new_x = {k: v[idx_to_take] for k, v in x.items()}
        else:
            new_x = x

        return new_x, y


class TrxDataset(Dataset):
    def __init__(self, data, y_dtype=np.float32, style='map', with_target=True):
        self.data = data
        self.y_dtype = y_dtype
        self.with_target = with_target

        if isinstance(data, torch.utils.data.IterableDataset):
            self.style = 'iterable'
        else:
            self.style = style

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        for rec in iter(self.data):
            x = rec['feature_arrays']
            if self.with_target:
                y = rec.get('target', None)
                yield x, self.y_dtype(y)
            else:
                yield x

    def __getitem__(self, idx):
        data = self.data[idx]
        x = data['feature_arrays']
        y = data.get('target', None)

        if self.with_target:
            return x, self.y_dtype(y)
        else:
            return x


class ConvertingTrxDataset(Dataset):
    def __init__(self, delegate, style='map', with_target=True):
        self.delegate = delegate
        self.with_target = with_target
        if hasattr(delegate, 'style'):
            self.style = delegate.style
        else:
            self.style = style

    def __len__(self):
        return len(self.delegate)

    def __iter__(self):
        for rec in iter(self.delegate):
            yield self._one_item(rec)

    def __getitem__(self, idx):
        item = self.delegate[idx]
        if type(item) is list:
            return [self._one_item(t) for t in item]
        else:
            return self._one_item(item)

    def _one_item(self, item):
        if self.with_target:
            x, y = item
            x = {k: torch.from_numpy(self.to_torch_compatible(v)) for k, v in x.items()}
            return x, y
        else:
            item = {k: torch.from_numpy(self.to_torch_compatible(v)) for k, v in item.items()}
            return item

    @staticmethod
    def to_torch_compatible(a):
        if a.dtype == np.int8:
            return a.astype(np.int16)
        return a


class ProcessDataset(Dataset):
    def __init__(self, delegate, process_fun, style='map'):
        self.delegate = delegate
        self.process_fun = process_fun
        if hasattr(delegate, 'style'):
            self.style = delegate.style
        else:
            self.style = style

    def __len__(self):
        return len(self.delegate)

    def __iter__(self):
        for rec in iter(self.delegate):
            yield self._one_item(rec)

    def __getitem__(self, idx):
        item = self.delegate[idx]
        if type(item) is list:
            return [self._one_item(t) for t in item]
        else:
            return self._one_item(item)

    def _one_item(self, item):
        x, y = item
        x = self.process_fun(x)
        return x, y


def pad_sequence(sequence, alignment, max_len=None, pad_value=0.0):
    def get_pad(x, max_len):
        if alignment == 'left':
            return max_len - len(x), 0
        if alignment == 'right':
            return 0, max_len - len(x)
        else:
            raise AttributeError(f"Unknown `alignment`: {alignment}")

    if max_len is None:
        max_len = max(x.size()[0] for x in sequence)
    return torch.stack([torch.nn.functional.pad(x, get_pad(x, max_len), value=pad_value) for x in sequence])


def padded_collate(batch):
    new_x_ = defaultdict(list)
    for x, _ in batch:
        for k, v in x.items():
            new_x_[k].append(v)

    lengths = torch.IntTensor([len(e) for e in next(iter(new_x_.values()))])

    new_x = {k: torch.nn.utils.rnn.pad_sequence(v, batch_first=True) for k, v in new_x_.items()}
    new_y = np.array([y for _, y in batch])
    if new_y.dtype.kind in ('i', 'f'):
        new_y = torch.from_numpy(new_y)

    return PaddedBatch(new_x, lengths), new_y


def padded_collate_distribution_target(batch):
    padded_batch, new_y = padded_collate(batch)

    keys = ['neg_sum', 'neg_distribution', 'pos_sum', 'pos_distribution']

    if new_y.shape[1] > 2:
        new_y = {key : new_y[:, i] for i, key in enumerate(keys)}
    else:
        new_y = {key : (new_y[:, i - 2] if i > 1 else 0) for i, key in enumerate(keys)}
    return padded_batch, new_y


def padded_collate_emb_valid(batch):
    x = batch[0]['prev_embeds'][None, :]
    neg_sum = []
    pos_sum = []
    neg_distribution = []
    pos_distribution = []
    for d in batch:
        neg_sum += [d['sum_n']]
        pos_sum += [d['sum_p']]
        neg_distribution += [d['share_n'].tolist()]
        pos_distribution += [d['share_p'].tolist()]
        x = torch.cat([x, d['prev_embeds'][None, :]])
    x = x[1:]
    neg_distribution += [[]]
    pos_distribution += [[]]
    neg_distribution = np.array(neg_distribution, dtype=object)
    pos_distribution = np.array(pos_distribution, dtype=object)
    neg_distribution = neg_distribution[:-1]
    pos_distribution = pos_distribution[:-1]
    neg_sum = np.array(neg_sum)
    pos_sum = np.array(pos_sum)
    y = {'neg_sum': neg_sum,
         'pos_sum': pos_sum,
         'neg_distribution': neg_distribution,
         'pos_distribution': pos_distribution}
    return x.float(), y


def padded_collate_wo_target(batch):
    new_x_ = defaultdict(list)
    for x in batch:
        for k, v in x.items():
            new_x_[k].append(v)

    lengths = torch.IntTensor([len(e) for e in next(iter(new_x_.values()))])
    new_x = {k: torch.nn.utils.rnn.pad_sequence(v, batch_first=True) for k, v in new_x_.items()}
    return PaddedBatch(new_x, lengths)
    

class ZeroDownSampler(Sampler):
    def __init__(self, targets):
        super().__init__(None)
        self.pos_ids = np.array([idx for idx, y in enumerate(targets) if y == 1])
        self.neg_ids = np.array([idx for idx, y in enumerate(targets) if y == 0])

    def __iter__(self):
        neg_idx = np.random.choice(len(self.neg_ids), size=min(len(self.pos_ids), len(self.neg_ids)), replace=False)
        pos_idx = np.random.choice(len(self.pos_ids), size=len(self.pos_ids), replace=False)

        sampler_order = np.concatenate((self.pos_ids[pos_idx], self.neg_ids[neg_idx]))
        np.random.shuffle(sampler_order)
        return iter(sampler_order)

    def __len__(self):
        return len(self.pos_ids)*2


def create_weighted_random_sampler(targets):
    n_samples = len(targets)
    n_pos = sum(targets)
    n_neg = n_samples - n_pos
    w_neg = np.min([n_pos / n_neg, 1.])

    weights = [1.0 if y == 1 else w_neg for y in targets]
    n_take = int(np.min([n_pos * 2, n_pos + n_neg]))
    return WeightedRandomSampler(weights, n_take)


def create_train_loader(dataset, params):
    if params.get('random_neg', False):
        targets = [y for x, y in dataset]
        sampler = ZeroDownSampler(targets)
    else:
        sampler = None

    dataset = DropoutTrxDataset(dataset, params.trx_dropout, params.max_seq_len)

    valid_loader = DataLoader(
        dataset,
        batch_size=params.batch_size,
        shuffle=sampler is None,
        sampler=sampler,
        num_workers=params.num_workers,
        collate_fn=padded_collate)

    return valid_loader


class MapStyleDatasetWrapper(torch.utils.data.Dataset):
    """Deprecated. Use ptls.data_load.datasets.memory_dataset.MemoryMapDataset
    """
    def __init__(self, data):
        self.data = data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]


class IterableDatasetWrapper(torch.utils.data.IterableDataset):
    """Deprecated. Use ptls.data_load.datasets.memory_dataset.MemoryIterableDataset
    """
    def __init__(self, data):
        warnings.warn('Multiply data in multiprocessing mode', UserWarning)
        self.data = data

    def __iter__(self):
        return iter(self.data)


def create_validation_loader(dataset, params):
    """Deprecated

    """
    dataset = DropoutTrxDataset(dataset, 0, params.max_seq_len)

    if dataset.style == 'iterable':
        dataset = IterableDatasetWrapper(dataset)
        logger.info('IterableDatasetWrapper used')
    else:
        dataset = MapStyleDatasetWrapper(dataset)
        logger.info('MapStyleDatasetWrapper used')

    valid_loader = DataLoader(
        dataset,
        batch_size=params.batch_size,
        shuffle=False,
        num_workers=params.num_workers,
        collate_fn=padded_collate)

    return valid_loader


# def augmentation_chain(*i_filters):
#     def _func(x):
#         for f in i_filters:
#             x = f(x)
#         return x
#     return _func


class augmentation_chain:
    def __init__(self, *i_filters):
        self.i_filters = i_filters

    def __call__(self, x):
        for f in self.i_filters:
            x = f(x)
        return x


class AugmentationChain:
    def __init__(self, f_augmentations):
        self.f_augmentations = f_augmentations if f_augmentations is not None else []

    def __call__(self, x):
        for f in self.f_augmentations:
            x = f(x)
        return x


class IterableAugmentations(IterableProcessingDataset):
    def __init__(self, a_chain):
        super().__init__()

        self.a_chain = a_chain

    def __iter__(self):
        for row in self._src:
            if type(row) is tuple:
                x, *targets = row
                x = self.a_chain(x)
                yield tuple([x, *targets])
            else:
                x = row
                x = self.a_chain(x)
                yield x


class IterableChain:
    def __init__(self, *i_filters):
        self.i_filters = i_filters

    def __call__(self, seq):
        for f in self.i_filters:
            # logger.debug(f'Applied {f} to {seq}')
            seq = f(seq)
        # logger.debug(f'Returned {seq}')
        return seq



from torch.utils.data.dataset import IterableDataset
from ptls.data_load.feature_dict import FeatureDict


class IterableProcessingDataset(FeatureDict, IterableDataset):
    def __init__(self):
        super().__init__()

        self._src = None

    def __call__(self, src):
        self._src = src
        return iter(self)

    def __iter__(self):
        """For record transformation. Redefine __iter__ for filter"""
        for rec in self._src:
            if type(rec) is tuple:
                features = rec[0]
                new_features = self.process(features)
                yield tuple([new_features, *rec[1:]])
            else:
                features = rec
                new_features = self.process(features)
                yield new_features

    def process(self, features):
        raise NotImplementedError()



import torch
from ptls.frames.supervised.metrics import BucketAccuracy, JSDiv, UnivMeanError, LogAccuracy, RankAUC
from ptls.loss import ZILNLoss


def test_bucket_accuracy():
    m, B = BucketAccuracy(), 100
    assert m(torch.randn(B), torch.randn(B)) >= 0
    assert m(torch.randn(B, 3), torch.randn(B)) >= 0
    assert m(torch.randn(B), torch.randn(B, 3)) >= 0
    assert m(torch.randn(B, 3), torch.randn(B, 3)) >= 0


def test_jsdiv():
    m = JSDiv()
    y = torch.zeros(10, 5)
    assert m(torch.randn(y.shape[0], y.shape[1] + 3), y) >= 0


def test_univ_mean_error():
    m, B = UnivMeanError(), 10
    assert m(torch.randn(B), torch.randn(B)) >= 0
    assert m(torch.randn(B, 3), torch.randn(B)) >= 0
    assert m(torch.randn(B, 3), torch.randn(B, 3)) >= 0


def test_accuracy_bin():
    acc = LogAccuracy()
    y_hat = torch.tensor([0.1, 0.4, 0.6, 0.8, 0.9])
    y = torch.tensor([0, 1, 0, 1, 0])
    acc(y_hat, y)
    assert acc.compute().mul(100).int() == 40


def test_accuracy_mul():
    acc = LogAccuracy()
    y_hat = torch.log_softmax(torch.tensor([
        [-1, 2, 1],
        [1, 2, -1],
        [1, -2, 0],
        [1, 1, 2],
        [1, 1, 2],
    ]).float(), dim=1)
    y = torch.tensor([0, 1, 1, 2, 2])
    acc(y_hat, y)
    assert acc.compute().mul(100).int() == 60


def test_ziln_loss():
    loss = ZILNLoss()
    min_loss = 0.5 * torch.log(torch.tensor(loss.eps))
    y = torch.zeros(10, 5)
    assert loss(torch.zeros(y.shape[0]), y) == 0
    assert loss(torch.zeros(y.shape[0], 2), y) >= min_loss
    assert loss(torch.zeros(y.shape[0], 3), y) >= min_loss
    assert loss(torch.zeros(y.shape[0], 3 + y.shape[1]), y) >= min_loss


def test_rank_auc():
    m = RankAUC()
    assert m(torch.tensor([0.1, 0.5, 0.7, 0.9]), torch.arange(4).float()) == 1
    assert m(torch.tensor([0.9, 0.7, 0.5, 0.1]), torch.arange(4).float()) == 0
    assert m(torch.randn(10), torch.randn(10)) >= 0



from ptls.data_load.iterable_processing.category_size_clip import CategorySizeClip
import numpy as np


def get_data_without_target():
    return [
        {
            'mcc': np.array([0, 1, 2, 5, 8, 9, 11, 15]),
            'currency': np.array([0, 1, 2, 5, 8, 9, 11, 15]),
        }
    ]


def get_data_with_target():
    return [
        ({
            'mcc': np.array([0, 1, 2, 5, 8, 9, 11, 15]),
            'currency': np.array([0, 1, 2, 5, 8, 9, 11, 15]),
        }, 1)
    ]


def test_max_clip_without_target():
    i_filter = CategorySizeClip({'mcc': 10, 'currency': 4})
    data = i_filter(get_data_without_target())
    data = [x for x in data][0]
    np.testing.assert_equal(data['mcc'], np.array([0, 1, 2, 5, 8, 9, 9, 9]))
    np.testing.assert_equal(data['currency'], np.array([0, 1, 2, 3, 3, 3, 3, 3]))


def test_max_clip_with_target():
    i_filter = CategorySizeClip({'mcc': 10, 'currency': 4})
    data = i_filter(get_data_with_target())
    data = [x for x in data][0][0]
    np.testing.assert_equal(data['mcc'], np.array([0, 1, 2, 5, 8, 9, 9, 9]))
    np.testing.assert_equal(data['currency'], np.array([0, 1, 2, 3, 3, 3, 3, 3]))


def test_value_clip_without_target():
    i_filter = CategorySizeClip({'mcc': 10, 'currency': 4}, replace_value=1)
    data = i_filter(get_data_without_target())
    data = [x for x in data][0]
    np.testing.assert_equal(data['mcc'], np.array([0, 1, 2, 5, 8, 9, 1, 1]))
    np.testing.assert_equal(data['currency'], np.array([0, 1, 2, 1, 1, 1, 1, 1]))


def test_value_clip_with_target():
    i_filter = CategorySizeClip({'mcc': 10, 'currency': 4}, replace_value=1)
    data = i_filter(get_data_with_target())
    data = [x for x in data][0][0]
    np.testing.assert_equal(data['mcc'], np.array([0, 1, 2, 5, 8, 9, 1, 1]))
    np.testing.assert_equal(data['currency'], np.array([0, 1, 2, 1, 1, 1, 1, 1]))




from ptls.frames.coles.coles_dataset import ColesDataset
from collections import defaultdict
import torch
from ptls.data_load.padded_batch import PaddedBatch


class CpcV2Dataset(ColesDataset):
    @staticmethod
    def collate_fn(batch):
        # TODO: refactoring required

        split_count = len(batch[0])

        sequences = [defaultdict(list) for _ in range(split_count)]
        lengths = torch.zeros((len(batch), split_count), dtype=torch.int)

        for i, client_data in enumerate(batch):
            for j, subseq in enumerate(client_data):
                for k, v in subseq.items():
                    sequences[j][k].append(v)
                    lengths[i][j] = v.shape[0]

        # adding padds in each split
        padded_batch = []
        for j, seq in enumerate(sequences):
            for k, v in seq.items():
                seq[k] = torch.nn.utils.rnn.pad_sequence(v, batch_first=True)
            padded_batch.append(PaddedBatch(seq, lengths[:, j].long()))

        return tuple(padded_batch)


class CpcV2IterableDataset(CpcV2Dataset, torch.utils.data.IterableDataset):
    pass



import numpy as np
import torch

from ptls.frames.coles.sampling_strategies.pair_selector import PairSelector
from ptls.frames.coles.metric import outer_pairwise_distance


class DistanceWeightedPairSelector(PairSelector):
    """
    Distance Weighted Sampling

    "Sampling Matters in Deep Embedding Learning", ICCV 2017
    https://arxiv.org/abs/1706.07567
    code based on https://github.com/suruoxi/DistanceWeightedSampling

    Generates pairs correspond to distances

    parameters
    ----------
    batch_k: int
        number of images per class

    Inputs:
        data: input tensor with shape (batch_size, embed_dim)
            Here we assume the consecutive batch_k examples are of the same class.
            For example, if batch_k = 5, the first 5 examples belong to the same class,
            6th-10th examples belong to another class, etc.
    Outputs:
        a_indices: indicess of anchors
        x[a_indices]
        x[p_indices]
        x[n_indices]
        xxx

    """

    def __init__(self, batch_k, cutoff=0.5, nonzero_loss_cutoff=1.4, normalize=False):
        super(DistanceWeightedPairSelector, self).__init__()
        self.batch_k = batch_k
        self.cutoff = cutoff
        self.nonzero_loss_cutoff = nonzero_loss_cutoff
        self.normalize = normalize

    def get_pairs(self, x, labels):
        k = self.batch_k
        n, d = x.shape
        distance = outer_pairwise_distance(x.detach())
        distance = distance.clamp(min=self.cutoff)
        log_weights = ((2.0 - float(d)) * distance.log() - (float(d - 3) / 2) * torch.log(
            torch.clamp(1.0 - 0.25 * (distance * distance), min=1e-8)))

        if self.normalize:
            log_weights = (log_weights - log_weights.min()) / (log_weights.max() - log_weights.min() + 1e-8)

        weights = torch.exp(log_weights - torch.max(log_weights))

        device = x.device
        weights = weights.to(device)

        mask = torch.ones_like(weights)
        for i in range(0, n, k):
            mask[i:i + k, i:i + k] = 0

        mask_uniform_probs = mask.double() * (1.0 / (n - k))

        weights = weights * mask * ((distance < self.nonzero_loss_cutoff).float()) + 1e-8
        weights_sum = torch.sum(weights, dim=1, keepdim=True)
        weights = weights / weights_sum

        a_indices = []
        p_indices = []
        n_indices = []

        np_weights = weights.cpu().numpy()
        for i in range(n):
            block_idx = i // k

            if weights_sum[i] != 0:
                n_indices += np.random.choice(n, k - 1, p=np_weights[i]).tolist()
            else:
                n_indices += np.random.choice(n, k - 1, p=mask_uniform_probs[i]).tolist()
            for j in range(block_idx * k, (block_idx + 1) * k):
                if j != i:
                    a_indices.append(i)
                    p_indices.append(j)

        positive_pairs = [[a, p] for a, p in zip(a_indices, p_indices)]
        negative_pairs = [[a, n] for a, n in zip(a_indices, n_indices)]

        return torch.LongTensor(positive_pairs).to(device), torch.LongTensor(negative_pairs).to(device)


import pandas as pd

from ptls.preprocessing.base.col_category_transformer import ColCategoryTransformer
from ptls.preprocessing.pandas.col_transformer import ColTransformerPandasMixin


class FrequencyEncoder(ColTransformerPandasMixin, ColCategoryTransformer):
    """Use frequency encoding for categorical field

    Let's `col_name_original` value_counts looks like this:
    cat value: records counts in dataset
          aaa:  100
          bbb:  50
          nan:  10
          ccc:  1

    Mapping will use this order to enumerate embedding indexes for category values:
    cat value: embedding id
    <padding token>: 0
                aaa: 1
                bbb: 2
                nan: 3
                ccc: 4
     <other values>: 5

    `dictionary_size` will be 6

    Parameters
    ----------
    col_name_original:
        Source column name
    col_name_target:
        Target column name. Transformed column will be placed here
        If `col_name_target is None` then original column will be replaced by transformed values.
    is_drop_original_col:
        When target and original columns are different manage original col deletion.
    """
    def __init__(self,
                 col_name_original: str,
                 col_name_target: str = None,
                 is_drop_original_col: bool = True,
                 ):
        super().__init__(
            col_name_original=col_name_original,
            col_name_target=col_name_target,
            is_drop_original_col=is_drop_original_col,
        )

        self.mapping = None
        self.other_values_code = None

    def fit(self, x: pd.DataFrame):
        super().fit(x)
        pd_col = x[self.col_name_original].astype(str)
        vc = pd_col.value_counts()
        self.mapping = {k: i + 1 for i, k in enumerate(vc.index)}
        self.other_values_code = len(vc) + 1
        return self

    @property
    def dictionary_size(self):
        return self.other_values_code + 1

    def transform(self, x: pd.DataFrame):
        pd_col = x[self.col_name_original].astype(str)
        x = self.attach_column(x, pd_col.map(self.mapping).fillna(self.other_values_code).rename(self.col_name_target))
        x = super().transform(x)
        return x



from ptls.data_load import augmentation_chain


def build_augmentations(conf):
    def _chain():
        from .all_time_shuffle import AllTimeShuffle
        from .dropout_trx import DropoutTrx
        from .random_slice import RandomSlice
        from .seq_len_limit import SeqLenLimit
        from .drop_day import DropDay

        for cls_name, params in conf:
            cls_f = locals().get(cls_name)
            if cls_f is None:
                raise AttributeError(f'Can not find augmentation for "{cls_name}"')
            yield cls_f(**params)

    return augmentation_chain(*_chain())



from .datasets import CpcDataset, CpcIterableDataset
from .datasets import CpcV2Dataset, CpcV2IterableDataset

from .modules.cpc_module import CpcModule
from .modules.cpc_v2_module import CpcV2Module



import torch
import numpy as np

from ptls.nn.normalization import L2NormEncoder


def test_norm_encoder():
    x = torch.tensor([
        [1.0, 0.0],
        [0.0, 2.0],
        [3.0, 4.0],
    ], dtype=torch.float64)

    f = L2NormEncoder()
    out = f(x).numpy()
    exp = np.array([
        [1.0, 0.0],
        [0.0, 1.0],
        [0.6, 0.8],
    ])
    np.testing.assert_array_almost_equal(exp, out)



import torch
import pytorch_lightning as pl
from ptls.data_load.padded_batch import PaddedBatch


class ABSModule(pl.LightningModule):
    @property
    def metric_name(self):
        raise NotImplementedError()

    @property
    def is_requires_reduced_sequence(self):
        raise NotImplementedError()

    def shared_step(self, x, y):
        """

        Args:
            x:
            y:

        Returns: y_h, y

        """
        raise NotImplementedError()

    def __init__(self, validation_metric=None,
                       seq_encoder=None,
                       loss=None,
                       optimizer_partial=None,
                       lr_scheduler_partial=None):
        """
        Parameters
        ----------
        params : dict
            params for creating an encoder
        seq_encoder : torch.nn.Module
            sequence encoder, if not provided, will be constructed from params
        """
        super().__init__()
        # self.save_hyperparameters()

        self._loss = loss
        self._seq_encoder = seq_encoder
        self._seq_encoder.is_reduce_sequence = self.is_requires_reduced_sequence
        self._validation_metric = validation_metric

        self._optimizer_partial = optimizer_partial
        self._lr_scheduler_partial = lr_scheduler_partial

    @property
    def seq_encoder(self):
        return self._seq_encoder

    def forward(self, x):
        return self._seq_encoder(x)

    def training_step(self, batch, _):
        y_h, y = self.shared_step(*batch)
        loss = self._loss(y_h, y)
        self.log('loss', loss)
        if type(batch) is tuple:
            x, y = batch
            if isinstance(x, PaddedBatch):
                self.log('seq_len', x.seq_lens.float().mean(), prog_bar=True)
        else:
            # this code should not be reached
            self.log('seq_len', -1, prog_bar=True)
            raise AssertionError('batch is not a tuple')
        return loss

    def validation_step(self, batch, _):
        y_h, y = self.shared_step(*batch)
        self._validation_metric(y_h, y)

    def validation_epoch_end(self, outputs):
        self.log(self.metric_name, self._validation_metric.compute(), prog_bar=True)
        self._validation_metric.reset()

    def configure_optimizers(self):
        optimizer = self._optimizer_partial(self.parameters())
        scheduler = self._lr_scheduler_partial(optimizer)
        
        if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
            scheduler = {
                'scheduler': scheduler,
                'monitor': self.metric_name,
            }
        return [optimizer], [scheduler]



import torch
import torchmetrics

from ptls.loss import BCELoss
from ptls.custom_layers import Squeeze
from ptls.frames.abs_module import ABSModule
from ptls.nn.seq_encoder.utils import AllStepsHead, FlattenHead


class RtdModule(ABSModule):
    """Replaced Token Detection (RTD) from [ELECTRA](https://arxiv.org/abs/2003.10555)

    Original sequence are encoded by `TrxEncoder`.
    Randomly sampled trx representations are replaced by an other trx representation from batch
    `seq_encoder` find replaced transactions.

    """
    def __init__(self, validation_metric=None,
                       seq_encoder=None,
                       head=None,
                       loss=None,
                       optimizer_partial=None,
                       lr_scheduler_partial=None):

        if validation_metric is None:
            validation_metric = torchmetrics.AUROC(num_classes=2)
        if loss is None:
            loss = BCELoss()

        super().__init__(validation_metric,
                         seq_encoder,
                         loss,
                         optimizer_partial,
                         lr_scheduler_partial)

        if head is None:
            self._head = torch.nn.Sequential(
                AllStepsHead(
                    torch.nn.Sequential(
                        torch.nn.Linear(self.seq_encoder.embedding_size, 1),
                        torch.nn.Sigmoid(),
                        Squeeze(),
                    )
                ),
                FlattenHead(),
            )
        else:
            self._head = head

    @property
    def metric_name(self):
        return 'valid_auroc'

    @property
    def is_requires_reduced_sequence(self):
        return False

    def shared_step(self, x, y):
        y_h = self.seq_encoder(x)
        y_h = self._head(y_h)
        return y_h, y



import numpy as np

from ptls.data_load.augmentations.dropout_trx import DropoutTrx


def test_no_dropout():
    i_filter = DropoutTrx(0.0)
    data = {'mcc': np.arange(100)}
    data = i_filter(data)
    assert len(data['mcc']) == 100
    assert (np.diff(data['mcc']) >= 0).all()


def test_with_dropout():
    i_filter = DropoutTrx(0.1)
    data = {'mcc': np.arange(100)}
    data = i_filter(data)
    assert len(data['mcc']) == 91
    assert (np.diff(data['mcc']) >= 0).all()

def test_with_target():
    i_filter = DropoutTrx(0.1)
    data = {'mcc': np.arange(100), 'target': 0}
    data = i_filter(data)
    assert len(data['mcc']) == 91
    assert (np.diff(data['mcc']) >= 0).all()



import numpy as np
import pandas as pd


def pd_hist(data, name, bins=10):
    if data.dtype.kind == 'f':
        bins = np.linspace(data.min(), data.max(), bins + 1).round(1)
    elif np.percentile(data, 99) - data.min() > bins - 1:
        bins = np.linspace(data.min(), np.percentile(data, 99), bins).astype(int).tolist() + [int(data.max() + 1)]
    else:
        bins = np.arange(data.min(), data.max() + 2, 1).astype(int)
    df = pd.cut(data, bins, right=False).rename(name)
    df = df.to_frame().assign(cnt=1).groupby(name)[['cnt']].sum()
    df['% of total'] = df['cnt'] / df['cnt'].sum()
    return df


def update_with_target(features, df_target, col_id, col_target):
    d_clients = df_target.to_dict(orient='index')
    features = [
        dict([('target', d_clients.get(rec[col_id], {}).get(col_target))] + list(rec.items()))
        for rec in features
    ]
    return features



import pytorch_lightning as pl
import torch
from pyhocon import ConfigFactory

from ptls.data_load import TrxDataset
from ptls.data_load import create_train_loader, create_validation_loader
from ptls.frames.cpc import CpcModule
from ptls_tests.utils.data_generation import gen_trx_data
from ptls.nn import RnnSeqEncoder
from ptls.nn import TrxEncoder
from functools import partial



def tst_params():
    params = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False
        },
        "trx_encoder": {
            "embeddings_noise": .1,
            "norm_embeddings": False,
            'embeddings': {
                'mcc_code': {'in': 21, 'out': 3},
                'trans_type': {'in': 11, 'out': 2},
            },
            'numeric_values': {'amount': 'log'}
        },
        "train": {
            "weight_decay": 0,
            "lr": 0.004,
            "batch_size": 32,
            "num_workers": 1,
            "trx_dropout": .1,
            "n_epoch": 1,
            "max_seq_len": 30,
        },
        "valid": {
            "batch_size": 32,
            "num_workers": 1,
            "max_seq_len": 30,
            "recall_top_k": 3
        },
    }

    params = ConfigFactory.from_dict(params)
    return params


def test_rnn_model():
    config = tst_params()

    pl_module = CpcModule(
        seq_encoder=RnnSeqEncoder(
            trx_encoder=TrxEncoder(**config['trx_encoder']),
            **config['rnn'],
        ),
        n_negatives=10, n_forward_steps=3,
        optimizer_partial=partial(torch.optim.Adam),
        lr_scheduler_partial=partial(torch.optim.lr_scheduler.StepLR, step_size=1, gamma=1.0),
    )
    train_data = gen_trx_data((torch.rand(1000)*60+1).long())
    train_ds = TrxDataset(train_data)
    test_data = gen_trx_data((torch.rand(100)*60+1).long())
    valid_ds = TrxDataset(test_data)

    train_loader = create_train_loader(train_ds, config['train'])
    valid_loader = create_validation_loader(valid_ds, config['valid'])

    trainer = pl.Trainer(
        gpus=None,
        max_steps=50,
        checkpoint_callback=False,
    )
    trainer.fit(pl_module, train_loader, valid_loader)



from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class IdFilterDf(IterableProcessingDataset):
    def __init__(self, df_relevant_ids):
        """Remove records which are not in `relevant_ids`
        Use possible multicolumn dataframe as `relevant_ids`

        Args:
            df_relevant_ids: list of ids which should be kept
        """
        super().__init__()

        self._df_relevant_ids = set(tuple(r) for r in df_relevant_ids.values.tolist())
        self.id_columns = df_relevant_ids.columns
        self.id_types = [type(v) for v in next(iter(self._df_relevant_ids))]

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            _id = tuple([col_type(features[col]) for col, col_type in zip(self.id_columns, self.id_types)])
            if _id not in self._df_relevant_ids:
                continue
            yield rec



from functools import reduce
from operator import iadd
from typing import List

from sklearn.base import BaseEstimator, TransformerMixin

from .col_category_transformer import ColCategoryTransformer
from .col_transformer import ColTransformer


class DataPreprocessor(BaseEstimator, TransformerMixin):
    def __init__(self,
                 ct_event_time: ColTransformer,
                 cts_category: List[ColCategoryTransformer],
                 cts_numerical: List[ColTransformer],
                 cols_identity: List[str],
                 t_user_group: ColTransformer,
                 ):
        self.ct_event_time = ct_event_time
        self.cts_category = cts_category
        self.cts_numerical = cts_numerical
        self.cols_identity = cols_identity
        self.t_user_group = t_user_group

        self._all_col_transformers = [
            [self.ct_event_time],
            self.cts_category,
            self.cts_numerical,
            [self.t_user_group],
        ]
        self._all_col_transformers = reduce(iadd, self._all_col_transformers, [])

    def fit(self, x):
        for i, ct in enumerate(self._all_col_transformers):
            if i == len(self._all_col_transformers):
                ct.fit(x)
            else:
                x = ct.fit_transform(x)
        return self

    def fit_transform(self, X, y=None, **fit_params):
        for ct in self._all_col_transformers:
            X = ct.fit_transform(X)
        return X

    def transform(self, x):
        for ct in self._all_col_transformers:
            x = ct.transform(x)
        return x

    def get_category_dictionary_sizes(self):
        """Gets a dict of mapping to integers lengths for categories
        """
        return {ct.col_name_target: ct.dictionary_size for ct in self.cts_category}

    def to_yaml(self):
        raise NotImplementedError()

    def from_yaml(self):
        raise NotImplementedError()



# coding: utf-8
"""

"""
import numpy as np


class AbsSplit:
    def split(self, dates):
        raise NotImplementedError()


class NoSplit(AbsSplit):
    def split(self, dates):
        date_len = dates.shape[0]
        date_range = np.arange(date_len)

        return [date_range]


class SampleRandom(AbsSplit):
    def __init__(self, split_count, cnt_min, cnt_max):
        self.split_count = split_count
        self.cnt_min = cnt_min
        self.cnt_max = cnt_max

    def split(self, dates):
        date_len = dates.shape[0]
        date_range = np.arange(date_len)

        lengths = np.random.randint(self.cnt_min, self.cnt_max, self.split_count)
        splits = []
        for i, l in enumerate(lengths):
            rand_perm = np.random.permutation(date_len)
            splits.append(np.sort(date_range[rand_perm][:l]))
        return splits


class SplitRandom(AbsSplit):
    def __init__(self, split_count, cnt_min, cnt_max):
        self.split_count = split_count
        self.cnt_min = cnt_min
        self.cnt_max = cnt_max

    def split(self, dates):
        date_len = dates.shape[0]
        date_range = np.arange(date_len)

        split_indexes = np.random.randint(0, self.split_count, date_len)
        return [date_range[split_indexes == i][:self.cnt_max] for i in range(self.split_count)]


class SplitByWeeks(AbsSplit):
    def __init__(self, split_count, cnt_max, seed=29, week_length=7, **kwargs):
        self.rs = np.random.RandomState(seed)
        self.split_count = split_count
        self.cnt_max = cnt_max
        self.week_length = week_length

    def split(self, dates):
        seq_len = len(dates)
        n_weeks = ((dates - dates[0]).astype('timedelta64[D]') // self.week_length).astype(np.int16)
        n_weeks_unique = np.unique(n_weeks)
        n_weeks_nunique = len(n_weeks_unique)

        if n_weeks_nunique < self.split_count:  # split by days
            n_weeks = (dates - dates[0]).astype(np.int16)
            n_weeks_unique = np.unique(n_weeks)
            n_weeks_nunique = len(n_weeks_unique)

        if n_weeks_nunique < self.split_count:  # split random
            n_weeks = np.arange(seq_len)
            n_weeks_unique = n_weeks
            n_weeks_nunique = len(n_weeks_unique)

        # divide n_weeks_unique for split_count parts
        split_indexes = np.random.randint(0, self.split_count, n_weeks_nunique - self.split_count)
        week_range = np.arange(self.split_count, n_weeks_nunique)
        n_weeks_idxes = [week_range[split_indexes == i] for i in range(self.split_count)]
        n_weeks_idxes = [np.append(x, i) for i, x in enumerate(n_weeks_idxes)]

        # select indexes correspond to each part of weeks
        # x_ij == 1 <=>  j-th elenemt of sequence correspond to n_weeks_unique[i]
        x = n_weeks.reshape(1, -1).repeat(n_weeks_nunique, axis=0) == n_weeks_unique.reshape(-1, 1).repeat(seq_len,
                                                                                                           axis=1)
        n_byweeks_idxes = [x[one_week_idxes].sum(axis=0).nonzero()[0] for one_week_idxes in n_weeks_idxes]
        n_byweeks_idxes = [x[-self.cnt_max:] if len(x) > self.cnt_max else x for x in n_byweeks_idxes]
        return n_byweeks_idxes


class SampleSlices(AbsSplit):
    def __init__(self, split_count, cnt_min, cnt_max, short_seq_crop_rate=1.0, is_sorted=False):
        self.split_count = split_count
        self.cnt_min = cnt_min
        self.cnt_max = cnt_max
        self.short_seq_crop_rate = short_seq_crop_rate
        self.is_sorted = is_sorted

    def split(self, dates):
        date_len = dates.shape[0]
        date_range = np.arange(date_len)

        if date_len <= self.cnt_min and self.short_seq_crop_rate >= 1.0:
            return [date_range for _ in range(self.split_count)]

        if int(date_len * self.short_seq_crop_rate) <= self.cnt_min and self.short_seq_crop_rate < 1.0:
            cnt_min = int(date_len * self.short_seq_crop_rate)
        else:
            cnt_min = self.cnt_min

        cnt_max = self.cnt_max if date_len > self.cnt_max else date_len

        lengths = np.random.randint(cnt_min, cnt_max+1, self.split_count)
        available_start_pos = (date_len - lengths).clip(0, None)
        start_pos = (np.random.rand(self.split_count) * (available_start_pos + 1 - 1e-9)).astype(int)
        if not self.is_sorted:
            return [date_range[s:s + l] for s, l in zip(start_pos, lengths)]

        ix_sort = np.argsort(start_pos)
        return [date_range[s:s + l] for s, l in zip(start_pos[ix_sort], lengths[ix_sort])]


class SampleUniform(AbsSplit):
    """
    Sub samples with equal length = `seq_len`
    Start pos has fixed uniform distribution from sequence start to end with equal step
    |---------------------|       main sequence
    |------|              |        sub seq 1
    |    |------|         |        sub seq 2
    |         |------|    |        sub seq 3
    |              |------|        sub seq 4

    There is no random factor in this splitter, so sub sequences are the same every time
    Can be used during inference as test time augmentation
    """
    def __init__(self, split_count, seq_len, **_):
        self.split_count = split_count
        self.seq_len = seq_len

    def split(self, dates):
        date_len = dates.shape[0]
        date_range = np.arange(date_len)

        if date_len <= self.seq_len + self.split_count:
            return [date_range for _ in range(self.split_count)]

        start_pos = np.linspace(0, date_len - self.seq_len, self.split_count).round().astype(int)
        return [date_range[s:s + self.seq_len] for s in start_pos]

class SampleUniformBySplitCount(AbsSplit):
    """
    Split into n sections:
    |---------------------|       main sequence
    |------|              |        sub seq 1
    |       |------|      |        sub seq 2
    |              |------|        sub seq 3
    There is no random factor in this splitter, so sub sequences are the same every time
    Can be used during inference as test time augmentation
    """
    def __init__(self, split_count, **_):
        self.split_count = split_count

    def split(self, dates):
        date_range = np.arange(dates.shape[0])
        return np.array_split(date_range, self.split_count)
        
class CutByDays(AbsSplit):
    def __init__(self, first_date, last_date):
        self.days_arange = np.arange(first_date, last_date+1)

    def split(self, dates):
        all_indexes = np.arange(len(dates))
        left_indexes = np.searchsorted(dates.astype(np.int32), self.days_arange)
        indexes = [all_indexes[:x] for x in left_indexes]
        return indexes


def create(split_strategy, **params):
    cls = globals().get(split_strategy, None)
    if cls is None:
        raise AttributeError(f'Unknown split_strategy: "{split_strategy}"')
    if not issubclass(cls, AbsSplit):
        raise AttributeError(f'Wrong split_strategy: "{split_strategy}". This is not a splitter')
    if cls is AbsSplit:
        raise AttributeError(f'Wrong split_strategy: "{split_strategy}". Splitter can not be abstract')

    return cls(**params)



import logging
from typing import List, Union

import numpy as np
import pandas as pd

from .base import DataPreprocessor, ColTransformer
from .base.col_category_transformer import ColCategoryTransformer
from .pandas.category_identity_encoder import CategoryIdentityEncoder
from .pandas.col_identity_transformer import ColIdentityEncoder
from .pandas.event_time import DatetimeToTimestamp
from .pandas.frequency_encoder import FrequencyEncoder
from .pandas.user_group_transformer import UserGroupTransformer

logger = logging.getLogger(__name__)


class PandasDataPreprocessor(DataPreprocessor):
    """Data preprocessor based on pandas.DataFrame

    During preprocessing it
        * transform datetime column to `event_time`
        * encodes category columns into indexes;
        * groups flat data by `col_id`;
        * arranges data into list of dicts with features

    Preprocessor don't modify original dataframe, but links to his data.

    Parameters
    ----------
    col_id : str
        name of column with ids. Used for groups
    col_event_time : str
        name of column with datetime
        or `ColTransformer` implementation with datetime transformation
    event_time_transformation: str
        name of transformation for `col_event_time`
        - 'dt_to_timestamp': datetime (string of datetime64) to timestamp (long) with `DatetimeToTimestamp`
            Original column is dropped by default cause target col `event_time` is the same information
            and we can not use as feature datetime column itself.
        - 'none': without transformation, `col_event_time` is in correct format. Used `ColIdentityEncoder`
            Original column is kept by default cause it can be any type and we may use it in the future
    cols_category : list[str]
        list of category columns. Each can me column name or `ColCategoryTransformer` implementation.
    category_transformation: str
        name of transformation for column names from `cols_category`
        - 'frequency': frequency encoding with `FrequencyEncoder`
        - 'none': no transformation with `CategoryIdentityEncoder`
    cols_numerical : list[str]
        list of columns to be mentioned as numerical features. No transformation with `ColIdentityEncoder`
    cols_identity : list[str]
        list of columns to be passed as is without any transformation
    cols_first_item: List[str]
        Only first value will be taken for these columns
        It can be user-level information joined to each transaction
    return_records:
        False: Result is a `pandas.DataFrame`.
            You can:
            - join any additional information like user-level features of target
            - convert it to `ptls` format using `.to_dict(orient='records')`
        True: Result is a list of dicts - `ptls` format

    """

    def __init__(self,
                 col_id: str,
                 col_event_time: Union[str, ColTransformer],
                 event_time_transformation: str = 'dt_to_timestamp',
                 cols_category: List[Union[str, ColCategoryTransformer]] = None,
                 category_transformation: str = 'frequency',
                 cols_numerical: List[str] = None,
                 cols_identity: List[str] = None,
                 cols_first_item: List[str] = None,
                 return_records: bool = True,
                 ):
        if cols_category is None:
            cols_category = []
        if cols_numerical is None:
            cols_numerical = []
        if cols_identity is None:
            cols_identity = []
        if cols_first_item is None:
            cols_first_item = []

        if type(col_event_time) is not str:
            ct_event_time = col_event_time  # use as is
        elif event_time_transformation == 'dt_to_timestamp':
            ct_event_time = DatetimeToTimestamp(col_name_original=col_event_time)
        elif event_time_transformation == 'none':
            ct_event_time = ColIdentityEncoder(
                col_name_original=col_event_time,
                col_name_target='event_time',
                is_drop_original_col=False,
            )
        else:
            raise AttributeError(f'incorrect event_time parameters combination: '
                                 f'`ct_event_time` = "{col_event_time}" '
                                 f'`event_time_transformation` = "{event_time_transformation}"')

        cts_category = []
        for col in cols_category:
            if type(col) is not str:
                cts_category.append(col)  # use as is
            elif category_transformation == 'frequency':
                cts_category.append(FrequencyEncoder(col_name_original=col))
            elif category_transformation == 'none':
                cts_category.append(CategoryIdentityEncoder(col_name_original=col))
            else:
                raise AttributeError(f'incorrect category parameters combination: '
                                     f'`cols_category[i]` = "{col}" '
                                     f'`category_transformation` = "{category_transformation}"')

        cts_numerical = [ColIdentityEncoder(col_name_original=col) for col in cols_numerical]
        t_user_group = UserGroupTransformer(
            col_name_original=col_id, cols_first_item=cols_first_item, return_records=return_records)

        super().__init__(
            ct_event_time=ct_event_time,
            cts_category=cts_category,
            cts_numerical=cts_numerical,
            cols_identity=cols_identity,
            t_user_group=t_user_group,
        )

    @staticmethod
    def _td_default(df, cols_event_time):
        df_event_time = df[cols_event_time].drop_duplicates()
        df_event_time = df_event_time.sort_values(cols_event_time)
        df_event_time['event_time'] = np.arange(len(df_event_time))
        df = pd.merge(df, df_event_time, on=cols_event_time)
        logger.info('Default time transformation')
        return df

    @staticmethod
    def _td_float(df, col_event_time):
        df['event_time'] = df[col_event_time].astype(float)
        logger.info('To-float time transformation')
        return df

    @staticmethod
    def _td_gender(df, col_event_time):
        """Gender-dataset-like transformation

        'd hh:mm:ss' -> float where integer part is day number and fractional part is seconds from day begin
        '1 00:00:00' -> 1.0
        '1 12:00:00' -> 1.5
        '1 01:00:00' -> 1 + 1 / 24
        '2 23:59:59' -> 1.99
        '432 12:00:00' -> 432.5

        :param df:
        :param col_event_time:
        :return:
        """
        padded_time = df[col_event_time].str.pad(15, 'left', '0')
        day_part = padded_time.str[:6].astype(float)
        time_part = pd.to_datetime(padded_time.str[7:], format='%H:%M:%S').values.astype(int) // 1e9
        time_part = time_part % (24 * 60 * 60) / (24 * 60 * 60)
        df['event_time'] = day_part + time_part
        logger.info('Gender-dataset-like time transformation')
        return df

    def _td_hours(self, df, col_event_time):
        logger.info('To hours time transformation')
        df['event_time'] = pd.to_datetime(df[col_event_time])
        df['event_time'] = (df['event_time'] - self.time_min).dt.total_seconds() / 3600

        return df



import numpy as np

from ptls.data_load.iterable_processing_dataset import IterableProcessingDataset


class FeatureBinScaler(IterableProcessingDataset):
    def __init__(self, feature_bins: dict, idx_starts_from=0):
        """ Apply binarization by given levels

        Params
        ------
            feature_bins: keys are feature names, values is list of bins boarders
        """
        super().__init__()

        self._feature_bins = {name:np.asarray(sorted(bins)) for name, bins in feature_bins.items()}
        self._idx_starts_from = idx_starts_from

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec

            # replace values by its nearest bin idx (idx starts from self._idx_starts_from)
            for name, bins in self._feature_bins.items():
                features[name] = self.find_bin(features[name], bins) + self._idx_starts_from
            yield rec

    @staticmethod
    def find_bin(col, bins):
        idx = np.abs(col.reshape(-1, 1) - bins).argmin(axis=1)
        return idx


import torchmetrics


class CpcAccuracy(torchmetrics.MeanMetric):
    def __init__(self, loss):
        super().__init__()

        self.loss = loss

    def update(self, *input):
        super().update(self.loss.cpc_accuracy(*input))



import torch

from ptls.nn.seq_encoder.rnn_encoder import RnnEncoder
from ptls.data_load.padded_batch import PaddedBatch


def get_data():
    return PaddedBatch(
        payload=torch.arange(4*5*8).view(4, 8, 5).float(),
        length=torch.tensor([4, 2, 6, 8])
    )


def test_example():
    model = RnnEncoder(
        input_size=5,
        hidden_size=6,
        is_reduce_sequence=False,
    )

    x = get_data()

    out = model(x)
    assert out.payload.shape == (4, 8, 6)


def test_last_step():
    model = RnnEncoder(
        input_size=5,
        hidden_size=6,
        is_reduce_sequence=True,
    )

    x = get_data()

    h = model(x)
    assert h.shape == (4, 6)



import logging
import os
import warnings
from glob import glob
from itertools import chain
from typing import Union, List

import numpy as np
import torch
from omegaconf import ListConfig

from ptls.data_load import read_pyarrow_file
from ptls.data_load import IterableChain

logger = logging.getLogger(__name__)


class ParquetFiles:
    """Helper file which search parquet files in specified path.

    Spark saves parquet files in a partitioned.
    This means that the output name given when saving is not a file, but a directory.
    `ParquetDataset` required file names so
    `ParquetFiles` is an adapter to scan spark output path and find parquet files.

    One single parquet file taken for example from pandas will be passed as is without directory scan.

    Parameters
    ----------
    file_path:
        one path or list of paths
    take_ixes:
        indexes for final file list selecting.
        Be careful using `take_ixes` cause file list is unsorted

    """
    def __init__(self, file_path: Union[str, List[str], ListConfig], take_ixes=None):
        if type(file_path) not in (list, ListConfig):
            file_path = [file_path]

        self._data_files = []
        for path in file_path:
            self._data_files.extend(self.scan_path(path))

        if take_ixes is not None:
            warnings.warn('Be careful using `take_ixes` with unsorted data file list')
            self._data_files = np.array(self._data_files)[take_ixes].tolist()

    @staticmethod
    def scan_path(path):
        if os.path.isfile(path):
            return [path]
        if os.path.isdir(path):
            file_path = os.path.join(path, '*.parquet')
            return glob(file_path)
        return []

    @property
    def data_files(self):
        return self._data_files


class ParquetDataset(torch.utils.data.IterableDataset):
    """Lazy (IterableDataset) load from parquet files

    File structure:
    *.parquet

    File structure example:
    data/
        part1.parquet
        part2.parquet
        ...

    Each file is read sequentially

    Parameters
    ----------
    data_files:
        ParquetFile object with list of files or just list of files
    post_processing:
        - deprecated, use i_filters
    i_filters:
        - list of `ptls.data_load.iterable_processing` filters
    shuffle_files:
        - shuffle data_files before reading when True.
    cache_schema:
        - dict schema (feature names) will be read once
    shuffle_seed:
        - random seed for shuffle_files

    """
    def __init__(self, data_files: Union[ParquetFiles, List[str]],
                 post_processing=None,
                 i_filters: List = None,
                 shuffle_files=False, cache_schema=True, shuffle_seed=42):
        if type(data_files) is ParquetFiles:
            self.data_files = data_files.data_files
        else:
            self.data_files = data_files
        if i_filters is not None:
            self.post_processing = IterableChain(*i_filters)
        else:
            self.post_processing = post_processing
        if post_processing is not None:
            warnings.warn('`post_processing` parameter is deprecated, use `i_filters`')
        self.shuffle_files = shuffle_files
        self.cache_schema = cache_schema
        self.shuffle_seed = shuffle_seed
        self.rs = None

        self._worker_id = None
        self._num_workers = None
        self._shuffle_seed = None
        self._schema = None

    def _get_my_files(self):
        files = [(i, name) for i, name in enumerate(self.data_files)]
        my_files = [name for i, name in files if i % self._num_workers == self._worker_id]

        return my_files

    def _init_worker(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:  # single-process data loading, return the full iterator
            self._worker_id = 0
            self._num_workers = 1
            self._shuffle_seed = self.shuffle_seed
        else:  # in a worker process
            self._worker_id = worker_info.id
            self._num_workers = worker_info.num_workers
            self._shuffle_seed = worker_info.seed
        logger.debug(f'Started [{self._worker_id:02d}/{self._num_workers:02d}]')

    def __iter__(self):
        self._init_worker()

        my_files = self._get_my_files()
        if self.shuffle_files:
            rs = np.random.RandomState(self._shuffle_seed % 2**32)
            rs.shuffle(my_files)

        logger.debug(f'Iter [{self._worker_id:02d}/{self._num_workers:02d}]: {my_files}')
        gen = chain(*[self.iter_file(name) for name in my_files])
        if self.post_processing is not None:
            gen = self.post_processing(gen)
        return gen

    def iter_file(self, file_name):
        """

        :param file_name:
        :return: [(customer_id, features)]
        """
        logger.debug(f'[{self._worker_id}/{self._num_workers}] Iter file "{file_name}"')
        for rec in read_pyarrow_file(file_name, use_threads=True):
            rec = {k: self.to_torch(v) for k, v in rec.items()}
            yield rec

    @staticmethod
    def to_torch(x):
        if type(x) is np.ndarray and x.dtype.kind in ('i', 'f'):
            return torch.from_numpy(x)
        return x



import logging
import warnings

import torch

from ptls.data_load import IterableChain
from ptls.data_load.iterable_processing.to_torch_tensor import ToTorch

logger = logging.getLogger(__name__)


class MemoryMapDataset(torch.utils.data.Dataset):
    def __init__(self, data, i_filters=None):
        if i_filters is None:
            self.processed_data = [rec for rec in data]
        else:
            post_processor_filter = IterableChain(*i_filters)
            self.processed_data = [rec for rec in post_processor_filter(data)]
        logger.info(f'Loaded {len(self.processed_data)} records')

    def __len__(self):
        return len(self.processed_data)

    def __getitem__(self, item):
        return self.processed_data[item]


class MemoryIterableDataset(torch.utils.data.IterableDataset):
    def __init__(self, data, i_filters=None):
        raise NotImplementedError()
        """Multiprocessing case aren't defined
        """
        if i_filters is None:
            i_filters = []
        self.data = data
        self.post_processor_filter = IterableChain(ToTorch(), *i_filters)

    def __iter__(self):
        for rec in self.post_processor_filter(self.data):
            yield rec




import torch
from torch import nn as nn
from torch.nn import functional as F


class BinomialDevianceLoss(nn.Module):
    """
    Binomial Deviance loss

    "Deep Metric Learning for Person Re-Identification", ICPR2014
    https://arxiv.org/abs/1407.4979
    """

    def __init__(self, pair_selector, alpha=1, beta=1, C=1):
        super(BinomialDevianceLoss, self).__init__()
        self.alpha = alpha
        self.beta = beta
        self.C = C
        self.pair_selector = pair_selector

    def forward(self, embeddings, target):

        positive_pairs, negative_pairs = self.pair_selector.get_pairs(embeddings, target)

        pos_pair_similarity = F.cosine_similarity(embeddings[positive_pairs[:, 0]], embeddings[positive_pairs[:, 1]], dim=1)
        neg_pair_similarity = F.cosine_similarity(embeddings[negative_pairs[:, 0]], embeddings[negative_pairs[:, 1]], dim=1)

        pos_loss = torch.mean(torch.log(1 + torch.exp(-self.alpha*(pos_pair_similarity - self.beta))))
        neg_loss = torch.mean(torch.log(1 + torch.exp(self.alpha*self.C*(neg_pair_similarity - self.beta))))

        res_loss = (pos_loss + neg_loss) * (len(target))

        return res_loss


import torch
from torch.nn import MultiheadAttention
from torch.nn.functional import relu, gelu

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.seq_encoder.abs_seq_encoder import AbsSeqEncoder


class Encoder(AbsSeqEncoder):
    """Custom builded transformer encoder

    Parameters
    ----------
    input_size:
        input embedding size.
        Equals intermediate and output layer size cause transformer don't change vector dimentions
    num_attention_heads:
        The number of heads in the multiheadattention models
    intermediate_size:
        The dimension of the feedforward network model
    num_hidden_layers:
        The number of sub-encoder-layers in the encoder
    attn_block_mode:
        default - without training connection weights 
        rezero - from (https://arxiv.org/pdf/2003.04887.pdf)
    self_attn_mode:
        quadratic - classic torch.nn.MultiheadAttention
        linear-flow - FlowFormer from (https://arxiv.org/pdf/2203.16194.pdf)
    aggregation_mode:
        Aggregation strategies - {mean, sum, max}
    layer_norm:
        None - default value, without layer normalization
        pre - before attn and before feedforward
        post - after attn and after feedforward
    is_reduce_sequence (bool):
        False - returns PaddedBatch with all transactions embeddings after aggregation step
        True - returns Tensor with all transactions embeddings after aggregation step
    """

    def __init__(self, 
                input_size: int,
                intermediate_size: int = 2048, 
                num_hidden_layers: int = 8,
                num_attention_heads: int = 8,
                attn_block_mode: str = 'rezero', 
                self_attn_mode: str = 'quadratic', 
                aggregation_mode: str = 'mean',
                layer_norm=None,
                is_reduce_sequence=True):
        super().__init__(is_reduce_sequence=is_reduce_sequence)
        
        self.transformer = torch.nn.Sequential(
            *[AttentionBlock(
                input_size, intermediate_size, attn_block_mode, self_attn_mode, layer_norm, num_attention_heads
            ) for _ in range(num_hidden_layers)]
        )
        
        self.aggregation = Aggregation(reduction=aggregation_mode)
        self.is_reduce_sequence = is_reduce_sequence

    def forward(self, x: PaddedBatch):
        out = self.transformer(x.payload)
        out = self.aggregation(out)
        if self.is_reduce_sequence:
            return out
        return PaddedBatch(out, x.seq_lens)

    @property
    def embedding_size(self):
        return self.hidden_size


class FlowAttention(torch.nn.Module):
    """FlowFormer from (https://arxiv.org/pdf/2203.16194.pdf)"""

    def __init__(self, input_dim, num_heads):
        super().__init__()
        embed_dim = input_dim
        assert embed_dim % num_heads == 0, "Embedding dimension must be 0 modulo number of heads."

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads

        self.qkv_proj = torch.nn.Linear(input_dim, 3*embed_dim)
        self.out_proj = torch.nn.Linear(embed_dim, input_dim)

        self._reset_parameters()

    def _reset_parameters(self):
        # Original Transformer initialization, see PyTorch documentation
        torch.nn.init.xavier_uniform_(self.qkv_proj.weight)
        self.qkv_proj.bias.data.fill_(0)
        torch.nn.init.xavier_uniform_(self.out_proj.weight)
        self.out_proj.bias.data.fill_(0)

    def forward(self, x):
        batch_size, seq_length, embed_dim = x.size()
        qkv = self.qkv_proj(x)

        # Separate Q, K, V from linear output
        qkv = qkv.reshape(batch_size, seq_length, self.num_heads, 3 * self.head_dim)
        q, k, v = qkv.chunk(3, dim=-1)   # [Batch, SeqLen, Head, Dims]

        q = torch.sigmoid(q)
        k = torch.sigmoid(k)

        i = torch.sum(q * torch.sum(k, dim=1, keepdim=True), dim=3, keepdim=True)
        o = torch.sum(k * torch.sum(q, dim=1, keepdim=True), dim=3, keepdim=True)

        i_hat = torch.sum(q * torch.sum(k / o, dim=1, keepdim=True), dim=3, keepdim=True)
        o_hat = torch.sum(k * torch.sum(q / i, dim=1, keepdim=True), dim=3, keepdim=True)

        r = torch.matmul(q / i, torch.matmul(k.transpose(-2, -1), v * torch.softmax(o_hat, dim=1))) * torch.sigmoid(i_hat)

        values = r.reshape(batch_size, seq_length, embed_dim)
        out = self.out_proj(values)
        
        return out

class MLP(torch.nn.Module):
    
    def __init__(self, n_in, n_hidden, n_out, depth=2):
        super().__init__()
        self.mlp = torch.nn.Sequential(
            torch.nn.Linear(n_in, n_hidden),
            torch.nn.GELU(),
            *[torch.nn.Sequential(
                torch.nn.Linear(n_hidden, n_hidden),
                torch.nn.GELU(),
            ) for _ in range(depth-1)],
            torch.nn.Linear(n_hidden, n_out)
        )
        
    def forward(self, X):
        return self.mlp(X)
    
class Attention(torch.nn.Module):
    
    def __init__(self, embed_dim, num_heads, self_attn):
        super().__init__()
        
        self.self_attn = self_attn
        if self_attn=="quadratic":
            self.attn = MultiheadAttention(embed_dim, num_heads, batch_first=True)
        elif self_attn=="linear-flow":
            self.attn = FlowAttention(embed_dim, num_heads)
        elif self_attn=="linear-cross":
            pass # TODO
        
    def forward(self, X):
        
        if self.self_attn=="quadratic":
            X, _ = self.attn(X, X, X)
        elif self.self_attn=="linear-flow":
            X = self.attn(X)
        elif self.self_attn=="linear-cross":
            pass
        
        return X
    
class AttentionBlock(torch.nn.Module):
    
    def __init__(self, embed_dim, mlp_hidden_dim, attn_block, self_attn, layer_norm, num_heads=4):
        super().__init__()
        
        self.attn_block = attn_block
        self.layer_norm = layer_norm
        
        if self.attn_block=="rezero":
            self.alpha_attn = torch.nn.Parameter(torch.normal(torch.tensor(0.), torch.tensor(1e-6))) 
            self.alpha_mlp = torch.nn.Parameter(torch.normal(torch.tensor(0.), torch.tensor(1e-6)))
        
        self.attn = Attention(embed_dim, num_heads, self_attn)
        self.linear1 = torch.nn.Linear(embed_dim, mlp_hidden_dim)
        self.linear2 = torch.nn.Linear(mlp_hidden_dim, embed_dim)
        
    def forward(self, X):
        
        if self.attn_block=="rezero":
            if self.layer_norm == 'pre':
                X = X / (X.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
            Z = X + self.alpha_attn * self.attn(X)
            if self.layer_norm:
                Z = Z / (Z.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
            Z = Z + self.alpha_mlp * self.linear2(gelu(self.linear1(Z)))
            if self.layer_norm == 'post':
                Z = Z / (Z.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
            X = X + Z # double residual
        
        else: # no-ln
            if self.layer_norm == 'pre':
                X = X / (X.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
            Z = X + self.attn(X)
            if self.layer_norm:
                Z = Z / (Z.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
            Z = Z + self.linear2(gelu(self.linear1(Z)))
            if self.layer_norm == 'post':
                Z = Z / (Z.pow(2).sum(dim=-1, keepdim=True) + 1e-9).pow(0.5)
            X = X + Z # double residual
    
        return X
    
class Aggregation(torch.nn.Module):
    
    def __init__(self, reduction="mean"):
        super().__init__()
        self.reduction = reduction
        
    def forward(self, X):
        
        if self.reduction=="mean":
            x = X.mean(dim=1)
        elif self.reduction=="sum":
            x = X.sum(dim=1)
        elif self.reduction=="max":
            x, _ = torch.max(x, dim=1)
        else:
            x = X[:, 0]
        return x


import torch

from ptls.nn.seq_encoder.rnn_encoder import RnnEncoder
from ptls.nn.seq_encoder.transformer_encoder import TransformerEncoder
from ptls.nn.seq_encoder.longformer_encoder import LongformerEncoder
from ptls.nn.seq_encoder.custom_encoder import Encoder

class SeqEncoderContainer(torch.nn.Module):
    """Base container class for Sequence encoder.
    Include `TrxEncoder` and `ptls.seq_encoder.abs_seq_encoder.AbsSeqEncoder` implementation

    Parameters
        trx_encoder:
            TrxEncoder object
        seq_encoder_cls:
            AbsSeqEncoder implementation class
        input_size:
            input_size parameter for seq_encoder_cls
            If None: input_size = trx_encoder.output_size
            Set input_size explicit or use None if your trx_encoder object has output_size attribute
        seq_encoder_params:
            dict with params for seq_encoder_cls initialisation
        is_reduce_sequence:
            False - returns PaddedBatch with all transactions embeddings
            True - returns one embedding for sequence based on CLS token

    """
    def __init__(self,
                 trx_encoder,
                 seq_encoder_cls,
                 input_size,
                 seq_encoder_params,
                 is_reduce_sequence=True,  # most frequent default behavior
                 ):
        super().__init__()

        self.trx_encoder = trx_encoder
        self.seq_encoder = seq_encoder_cls(
            input_size=input_size if input_size is not None else trx_encoder.output_size,
            is_reduce_sequence=is_reduce_sequence,
            **seq_encoder_params,
        )

    @property
    def is_reduce_sequence(self):
        return self.seq_encoder.is_reduce_sequence

    @is_reduce_sequence.setter
    def is_reduce_sequence(self, value):
        self.seq_encoder.is_reduce_sequence = value

    @property
    def category_max_size(self):
        return self.trx_encoder.category_max_size

    @property
    def category_names(self):
        return self.trx_encoder.category_names

    @property
    def embedding_size(self):
        return self.seq_encoder.embedding_size

    def forward(self, x):
        x = self.trx_encoder(x)
        x = self.seq_encoder(x)
        return x


class RnnSeqEncoder(SeqEncoderContainer):
    """SeqEncoderContainer with RnnEncoder

    Supports incremental embedding calculation.
    Each RNN step requires previous hidden state. Hidden state passed through the iterations during sequence processing.
    Starting hidden state required by RNN. Starting hidden state are depends on `RnnEncoder.trainable_starter`.
    You can also provide starting hidden state to `forward` method as `h_0`.
    This can be useful when you need to `update` your embedding with new transactions.

    Example:
        >>> seq_encoder = RnnSeqEncoder(...)
        >>> embedding_0 = seq_encoder(data_0)
        >>> embedding_1 = seq_encoder(data_1, h_0=embedding_0)
        >>> embedding_2a = seq_encoder(data_2, h_0=embedding_1)
        >>> embedding_2b = seq_encoder(data_2)
        >>> embedding_2c = seq_encoder(data_0 + data_1 + data_2)

    `embedding_2a` takes into account all transactions from `data_0`, `data_1` and `data_2`.
    `embedding_2b` takes into account only transactions from `data_2`.
    `embedding_2c` is the same as `embedding_2a`.
    `embedding_2a` calculated faster than `embedding_2c`.

    Incremental calculation works fast when you have long sequences and short updates. RNN just process short update.


    Parameters
        trx_encoder:
            TrxEncoder object
        input_size:
            input_size parameter for RnnEncoder
            If None: input_size = trx_encoder.output_size
            Set input_size explicit or use None if your trx_encoder object has output_size attribute
        **seq_encoder_params:
            params for RnnEncoder initialisation
        is_reduce_sequence:
            False - returns PaddedBatch with all transactions embeddings
            True - returns one embedding for sequence based on CLS token

    """
    def __init__(self,
                 trx_encoder=None,
                 input_size=None,
                 is_reduce_sequence=True,
                 **seq_encoder_params,
                 ):
        super().__init__(
            trx_encoder=trx_encoder,
            seq_encoder_cls=RnnEncoder,
            input_size=input_size,
            seq_encoder_params=seq_encoder_params,
            is_reduce_sequence=is_reduce_sequence,
        )

    def forward(self, x, h_0=None):
        x = self.trx_encoder(x)
        x = self.seq_encoder(x, h_0)
        return x


class TransformerSeqEncoder(SeqEncoderContainer):
    """SeqEncoderContainer with TransformerEncoder

    Parameters
        trx_encoder:
            TrxEncoder object
        input_size:
            input_size parameter for TransformerEncoder
            If None: input_size = trx_encoder.output_size
            Set input_size explicit or use None if your trx_encoder object has output_size attribute
        **seq_encoder_params:
            params for TransformerEncoder initialisation
        is_reduce_sequence:
            False - returns PaddedBatch with all transactions embeddings
            True - returns one embedding for sequence based on CLS token

    """

    def __init__(self,
                 trx_encoder=None,
                 input_size=None,
                 is_reduce_sequence=True,
                 **seq_encoder_params,
                 ):
        super().__init__(
            trx_encoder=trx_encoder,
            seq_encoder_cls=TransformerEncoder,
            input_size=input_size,
            seq_encoder_params=seq_encoder_params,
            is_reduce_sequence=is_reduce_sequence,
        )


class LongformerSeqEncoder(SeqEncoderContainer):
    """SeqEncoderContainer with TransformerEncoder

    Parameters
        trx_encoder:
            TrxEncoder object
        input_size:
            input_size parameter for LongformerEncoder
            If None: input_size = trx_encoder.output_size
            Set input_size explicit or use None if your trx_encoder object has output_size attribute
        **seq_encoder_params:
            params for LongformerEncoder initialisation
        is_reduce_sequence:
            False - returns PaddedBatch with all transactions embeddings
            True - returns one embedding for sequence based on CLS token

    """

    def __init__(self,
                 trx_encoder=None,
                 input_size=None,
                 is_reduce_sequence=True,
                 **seq_encoder_params,
                 ):
        super().__init__(
            trx_encoder=trx_encoder,
            seq_encoder_cls=LongformerEncoder,
            input_size=input_size,
            seq_encoder_params=seq_encoder_params,
            is_reduce_sequence=is_reduce_sequence,
        )

class CustomSeqEncoder(SeqEncoderContainer):
    """SeqEncoderContainer with Custom builded transformer encoder

    Parameters
        trx_encoder:
            TrxEncoder object
        input_size:
            input_size parameter for Custom Encoder
            If None: input_size = trx_encoder.output_size
            Set input_size explicit or use None if your trx_encoder object has output_size attribute
        **seq_encoder_params:
            params for for Custom Encoder initialisation
        is_reduce_sequence (bool):
            False - returns PaddedBatch with all transactions embeddings after aggregation step
            True - returns Tensor with all transactions embeddings after aggregation step

    """
    def __init__(self,
                 trx_encoder=None,
                 input_size=None,
                 is_reduce_sequence=True,
                 **seq_encoder_params,
                 ):
        super().__init__(
            trx_encoder=trx_encoder,
            seq_encoder_cls=Encoder,
            input_size=input_size,
            seq_encoder_params=seq_encoder_params,
            is_reduce_sequence=is_reduce_sequence,
        )

    def forward(self, x):
        z_trx = self.trx_encoder(x)
        out = self.seq_encoder(z_trx)
        return out


from functools import reduce
from operator import iadd

import torch

from ptls.data_load.feature_dict import FeatureDict
from ptls.data_load.utils import collate_feature_dict
from ptls.frames.coles.split_strategy import AbsSplit


class ColesDataset(FeatureDict, torch.utils.data.Dataset):
    """Dataset for ptls.frames.coles.CoLESModule

    Parameters
    ----------
    data:
        source data with feature dicts
    splitter:
        object from from `ptls.frames.coles.split_strategy`.
        Used to split original sequence into subsequences which are samples from one client.
    col_time:
        column name with event_time
    """

    def __init__(self,
                 data,
                 splitter: AbsSplit,
                 col_time='event_time',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)  # required for mixin class

        self.data = data
        self.splitter = splitter
        self.col_time = col_time

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        feature_arrays = self.data[idx]
        return self.get_splits(feature_arrays)

    def __iter__(self):
        for feature_arrays in self.data:
            yield self.get_splits(feature_arrays)

    def get_splits(self, feature_arrays):
        local_date = feature_arrays[self.col_time]
        indexes = self.splitter.split(local_date)
        return [{k: v[ix] for k, v in feature_arrays.items() if self.is_seq_feature(k, v)} for ix in indexes]

    @staticmethod
    def collate_fn(batch):
        class_labels = [i for i, class_samples in enumerate(batch) for _ in class_samples]
        batch = reduce(iadd, batch)
        padded_batch = collate_feature_dict(batch)
        return padded_batch, torch.LongTensor(class_labels)


class ColesIterableDataset(ColesDataset, torch.utils.data.IterableDataset):
    pass



import math

import torch
from torch import nn as nn


class FloatPositionalEncoding(nn.Module):
    def __init__(self, out_size):
        super(FloatPositionalEncoding, self).__init__()

        self.out_size = out_size

    def forward(self, position):
        """

        :param position: B x T
        :return: B x T x H
        """
        div_term = torch.exp(torch.arange(0, self.out_size, 2).float() * (-math.log(10000.0) / self.out_size))
        div_term = div_term.unsqueeze(0).unsqueeze(0)
        div_term = div_term.to(device=position.device)

        position = position.unsqueeze(2)

        pe = torch.cat([torch.sin(position * div_term), torch.cos(position * div_term)], dim=2)
        self.register_buffer('pe', pe)

        return pe



import torch


class BarlowTwinsLoss(torch.nn.Module):
    """
    From https://github.com/facebookresearch/barlowtwins

    """
    def __init__(self, lambd):
        super().__init__()

        self.lambd = lambd

    def forward(self, model_outputs, target):
        n = len(model_outputs)
        ix1 = torch.arange(0, n, 2, device=model_outputs.device)
        ix2 = torch.arange(1, n, 2, device=model_outputs.device)

        assert (target[ix1] == target[ix2]).all(), "Wrong embedding positions"

        z1 = model_outputs[ix1]
        z2 = model_outputs[ix2]

        # empirical cross-correlation matrix
        c = torch.mm(z1.T, z2)
        c.div_(n // 2)

        on_diag = torch.diagonal(c).add_(-1).pow_(2).sum()
        off_diag = self.off_diagonal(c).pow_(2).sum()
        loss = on_diag + self.lambd * off_diag
        return loss

    @staticmethod
    def off_diagonal(x):
        # return a flattened view of the off-diagonal elements of a square matrix
        n, m = x.shape
        assert n == m
        return x.flatten()[:-1].view(n - 1, n + 1)[:, 1:].flatten()


import torch

from ptls.frames.coles.sampling_strategies.pair_selector import PairSelector
from ptls.frames.coles.metric import outer_pairwise_distance


class HardNegativePairSelector(PairSelector):
    """
    Generates all possible possitive pairs given labels and
         neg_count hardest negative example for each example
    """

    def __init__(self, neg_count=1):
        super(HardNegativePairSelector, self).__init__()
        self.neg_count = neg_count

    def get_pairs(self, embeddings, labels):
        # construct matrix x, such as x_ij == 0 <==> labels[i] == labels[j]
        n = labels.size(0)
        x = labels.expand(n, n) - labels.expand(n, n).t()

        # positive pairs
        positive_pairs = torch.triu((x == 0).int(), diagonal=1).nonzero(as_tuple=False)

        # hard negative minning
        mat_distances = outer_pairwise_distance(embeddings.detach())  # pairwise_distance

        upper_bound = int((2 * n) ** 0.5) + 1
        mat_distances = ((upper_bound - mat_distances) * (x != 0).type(
            mat_distances.dtype))  # filter: get only negative pairs

        values, indices = mat_distances.topk(k=self.neg_count, dim=0, largest=True)
        negative_pairs = torch.stack([
            torch.arange(0, n, dtype=indices.dtype, device=indices.device).repeat(self.neg_count),
            torch.cat(indices.unbind(dim=0))
        ]).t()

        return positive_pairs, negative_pairs


import torch
from torch import nn


class BaseScaler(nn.Module):
    def __init__(self, col_name=None):
        super().__init__()
        self.col_name = col_name

    @property
    def output_size(self):
        raise NotImplementedError()


class IdentityScaler(BaseScaler):
    def forward(self, x):
        return x

    @property
    def output_size(self):
        return 1


class LogScaler(BaseScaler):
    def forward(self, x):
        return x.abs().log1p() * x.sign()

    @property
    def output_size(self):
        return 1


class YearScaler(BaseScaler):
    def forward(self, x):
        return x/365

    @property
    def output_size(self):
        return 1


class NumToVector(BaseScaler):
    def __init__(self, embeddings_size):
        super().__init__()
        self.w = torch.nn.Parameter(torch.randn(1, 1, embeddings_size), requires_grad=True)
        self.b = torch.nn.Parameter(torch.randn(1, 1, embeddings_size), requires_grad=True)

    def forward(self, x):
        return x * self.w + self.b

    @property
    def output_size(self):
        return self.w.size(2)


class LogNumToVector(BaseScaler):
    def __init__(self, embeddings_size):
        super().__init__()
        self.w = torch.nn.Parameter(torch.randn(1, 1, embeddings_size), requires_grad=True)
        self.b = torch.nn.Parameter(torch.randn(1, 1, embeddings_size), requires_grad=True)

    def forward(self, x):
        return x.abs().log1p() * x.sign() * self.w + self.b

    @property
    def output_size(self):
        return self.w.size(2)


def scaler_by_name(name):
    scaler = {
        'identity': IdentityScaler,
        'sigmoid': torch.nn.Sigmoid,
        'log': LogScaler,
        'year': YearScaler,
    }.get(name, None)

    if scaler is None:
        raise Exception(f'unknown scaler name: {name}')
    else:
        return scaler()


class PoissonScaler(BaseScaler):
    """
    Explicit estimator for poissonian target with standard pytorch sampler extrapolation.
    """
    def __init__(self, kmax=33):
        super().__init__()
        self.kmax = 0.7 * kmax
        self.arange = torch.nn.Parameter(torch.arange(kmax), requires_grad=False)
        self.factor = torch.nn.Parameter(torch.special.gammaln(1 + self.arange), requires_grad=False)

    def forward(self, x):
        if self.kmax == 0:
            return torch.poisson(x)
        res = self.arange * torch.log(x).unsqueeze(-1) - self.factor * torch.ones_like(x).unsqueeze(-1)
        return res.argmax(dim=-1).float().where(x < self.kmax, torch.poisson(x))

    @property
    def output_size(self):
        return 1


class ExpScaler(BaseScaler):
    def __init__(self, column=0):
        super().__init__()
        self.column = column

    def forward(self, x):
        if self.column is not None:
            return torch.exp(x if x.dim() == 1 else x[:, self.column].unsqueeze(-1))
        else:
            return torch.exp(x)

    @property
    def output_size(self):
        return 1



import logging

import numpy as np
import torch
import torch.nn.functional as F
import torchmetrics
from torch.special import entr

from ptls.loss import cross_entropy, kl, mape_metric, mse_loss, r_squared

logger = logging.getLogger(__name__)


class DistributionTargets(torchmetrics.Metric):
    def __init__(self, col_name):
        super().__init__()

        self.add_state('y_hat', default=[])
        self.add_state('y', default=[])
        self.col_name = col_name
        self.sign = -1 if self.col_name == 'neg_sum' else 1

    def update(self, y_hat, y):
        y_hat = y_hat[self.col_name]
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        y = torch.tensor(np.array(y[self.col_name].tolist(), dtype='float64'), device=device)
        self.y_hat.append(y_hat)
        self.y.append(y)


class CrossEntropy(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return cross_entropy(y_hat, y)


class KL(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return kl(y_hat, y)


class MSE(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return mse_loss(y_hat, torch.log(self.sign * y[:, None] + 1))


class MAPE(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return mape_metric(self.sign * torch.exp(y_hat - 1), y[:, None])


class LogAccuracy(torchmetrics.Metric):
    def __init__(self, **params):
        super().__init__(**params)

        self.add_state('correct', default=torch.tensor(0.0))
        self.add_state('total', default=torch.tensor(0.0))

    def update(self, preds, target):
        if len(preds.shape) == 1:
            preds = (preds > 0.5).to(dtype=target.dtype)
        else:
            preds = torch.argmax(preds, dim=1)
        self.correct += torch.sum(preds == target)
        self.total += target.numel()

    def compute(self):
        return self.correct / self.total


class R_squared(DistributionTargets):
    def __init__(self, col_name):
        super().__init__(col_name)

    def compute(self):
        y_hat = torch.cat(self.y_hat)
        y = torch.cat(self.y)
        return r_squared(self.sign * torch.exp(y_hat - 1), y[:, None])


class UnivMeanError(torchmetrics.Metric):
    """
    RMSE/MAE for multiple outputs with optional scaler.
    """
    is_differentiable = True
    higher_is_better = False
    full_state_update = False

    def __init__(self, calc_mae=True, scaler=None):
        super().__init__()
        self.calc_mae = calc_mae
        self.scaler = scaler
        self.add_state("psum", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("pnum", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, y1, y):
        t = y if y.dim() == 1 else y.sum(dim=1)
        if self.scaler is None:
            t1 = y1 if y1.dim() == 1 else y1.sum(dim=1)
        else:
            t1 = self.scaler(y1) if y1.dim() == 1 else self.scaler(y1).sum(dim=1)
        self.psum += torch.sum((t1 - t).abs() if self.calc_mae else (t1 - t).square())
        self.pnum += y.shape[0]

    def compute(self):
        res = self.psum / self.pnum
        return res if self.calc_mae else res.sqrt()


class BucketAccuracy(torchmetrics.Metric):
    """
    Multiclass Accuracy for bucketized regression variable.
    """
    is_differentiable = False
    higher_is_better = True
    full_state_update = False

    def __init__(self, n_buckets=10, scaler=None, compute_on_cpu=True):
        super().__init__(compute_on_cpu=compute_on_cpu)
        self.n_buckets = n_buckets
        self.scaler = scaler
        self.add_state("y1", default=[])
        self.add_state("y", default=[])

    def update(self, y1, y):
        if self.scaler is None:
            self.y1.append(y1 if y1.dim() == 1 else y1.sum(dim=1))
        else:
            self.y1.append(self.scaler(y1) if y1.dim() == 1 else self.scaler(y1).sum(dim=1))
        self.y.append(y if y.dim() == 1 else y.sum(dim=1))

    def compute(self):
        y1, y = torch.cat(self.y1).float().cpu(), torch.cat(self.y).float().cpu()
        q = torch.linspace(0, 1, self.n_buckets + 1).float().cpu()[1:]
        b1 = torch.quantile(y1, q, interpolation="nearest")
        b = torch.quantile(y, q, interpolation="nearest")
        y1 = torch.bucketize(y1, b1, out_int32=True)
        y = torch.bucketize(y, b, out_int32=True)
        return torchmetrics.functional.accuracy(y1, y)


class JSDiv(torchmetrics.Metric):
    """
    Jensen-Shannon divergence for distribution-like target:
        0 <= JSD(P||Q) = H(P/2 + Q/2) - H(P)/2 - H(Q)/2 <= ln(2)
    """
    is_differentiable = True
    higher_is_better = False
    full_state_update = False

    def __init__(self, scaler=None):
        super().__init__()
        self.scaler = scaler
        self.add_state("psum", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("pnum", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(self, y1, y):
        p = torch.cat(((y.sum(dim=1) == 0).float().unsqueeze(-1), y), dim=1)
        p /= p.sum(dim=1).unsqueeze(-1)
        if y1.shape[1] == y.shape[1] and self.scaler is not None:
            q = self.scaler(y1)
            q = torch.cat(((q.sum(dim=1) == 0).float().unsqueeze(-1), q), dim=1)
            q = q / q.sum(dim=1).unsqueeze(-1)
        elif y1.shape[1] == y.shape[1] + 1:
            q = F.softmax(y1, dim=1)
        elif y1.shape[1] == y.shape[1] + 3:
            q = F.softmax(y1[:, 2:], dim=1)
        else:
            raise Exception(f"{self.__class__} got incorrect input sizes")
        self.psum += torch.sum(entr(0.5 * (p + q)) - 0.5 * (entr(p) + entr(q)))
        self.pnum += y.shape[0]

    def compute(self):
        return self.psum / self.pnum


class RankAUC(torchmetrics.Metric):
    """
    Universal definition for AUC-ROC as ranking metric.
    """
    is_differentiable = False
    higher_is_better = True
    full_state_update = False

    def __init__(self, scaler=None, compute_on_cpu=True, max_size=22222):
        super().__init__(compute_on_cpu=compute_on_cpu)
        self.max_size = max_size
        self.scaler = scaler
        self.add_state("a", default=[])
        self.add_state("y", default=[])

    def update(self, a, y):
        if self.scaler is None:
            self.a.append(a if a.dim() == 1 else a.sum(dim=1))
        else:
            self.a.append(self.scaler(a) if a.dim() == 1 else self.scaler(a).sum(dim=1))
        self.y.append(y if y.dim() == 1 else y.sum(dim=1))

    def compute(self):
        a, y = torch.cat(self.a).float().cpu(), torch.cat(self.y).float().cpu()
        if a.shape[0] > self.max_size:
            sample_idx = torch.randperm(a.shape[0])[:self.max_size]
            a, y = a[sample_idx], y[sample_idx]
        ra = torch.heaviside(a.unsqueeze(0) - a.unsqueeze(1), torch.tensor([0.5]))
        ry = torch.heaviside(y.unsqueeze(0) - y.unsqueeze(1), torch.tensor([0.0]))
        return (ry * ra).sum() / ry.sum()



import pytest
import torch
from omegaconf import OmegaConf

from ptls.nn.seq_encoder.rnn_encoder import RnnEncoder
from ptls.nn.seq_encoder.utils import PerTransHead, scoring_head
from ptls.data_load.padded_batch import PaddedBatch


class TrxEncoderTest(torch.nn.Module):
    def forward(self, x):
        return x


# TODO: replace these tests with new-style usage patterns
def tst_rnn_model(config):
    p = TrxEncoderTest()
    e = RnnEncoder(8, **config['rnn'])
    h = scoring_head(16, config['head'])
    m = torch.nn.Sequential(p, e, h)
    return m


def tst_trx_avg_model():
    p = TrxEncoderTest()
    h = PerTransHead(8)
    m = torch.nn.Sequential(p, h)
    return m


def test_simple_config():
    config = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': False,
            'pred_all_states': False,
            'pred_all_states_mean': False,
            'norm_input': False,
            'use_batch_norm': False,
        }
    }
    config = OmegaConf.create(config)

    m = tst_rnn_model(config)

    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)
    out = m(x)

    assert len(out.size()) == 1
    assert out.size()[0] == 12


def test_concat_lens():
    config = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': True,
            'pred_all_states': False,
            'pred_all_states_mean': False,
            'norm_input': False,
            'use_batch_norm': False,
        },
    }
    config = OmegaConf.create(config)
    rnn = tst_rnn_model(config)

    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)
    out = rnn(x)

    assert len(out.size()) == 1
    assert out.size()[0] == 12


def test_trainable_starter():
    config = {
        'rnn': {
            'trainable_starter': 'static',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': True,
            'pred_all_states': False,
            'pred_all_states_mean': False,
            'norm_input': False,
            'use_batch_norm': False,
        },
    }
    config = OmegaConf.create(config)
    rnn = tst_rnn_model(config)

    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)
    out = rnn(x)

    assert len(out.size()) == 1
    assert out.size()[0] == 12


def test_rnn_type():
    config = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': False,
            'pred_all_states': False,
            'pred_all_states_mean': False,
            'norm_input': False,
            'use_batch_norm': False,
        },
    }
    config = OmegaConf.create(config)
    rnn = tst_rnn_model(config)

    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)
    out = rnn(x)

    assert len(out.size()) == 1
    assert out.size()[0] == 12


def test_pred_all_states():
    config = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': False,
            'pred_all_states': True,
            'pred_all_states_mean': False,
            'norm_input': False,
            'use_batch_norm': False,
        },
    }
    config = OmegaConf.create(config)

    rnn = tst_rnn_model(config)

    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)
    out = rnn(x).payload

    assert len(out.size()) == 2
    assert out.size()[0] == 12


def test_pred_all_states_mean():
    config = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': False,
            'pred_all_states': True,
            'pred_all_states_mean': True,
            'norm_input': False,
            'use_batch_norm': False,
        },
    }
    config = OmegaConf.create(config)
    rnn = tst_rnn_model(config)

    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)
    out = rnn(x)

    assert len(out.size()) == 1
    assert out.size()[0] == 12


def test_pred_all_states_no_effect():
    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)

    config = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': False,
            'pred_all_states': False,
            'pred_all_states_mean': False,
            'norm_input': False,
            'use_batch_norm': False,
        },
    }
    config = OmegaConf.create(config)
    rnn = tst_rnn_model(config)
    out1 = rnn(x)

    rnn.pred_all_states = True
    out2 = rnn(x)

    assert (out1 == out2).all()


def test_pred_all_states_and_concat_lens():
    config = {
        'rnn': {
            'trainable_starter': 'empty',
            'hidden_size': 16,
            'type': 'gru',
            'bidir': False,
        },
        'head': {
            'explicit_lengths': True,
            'pred_all_states': True,
            'pred_all_states_mean': False,
            'norm_input': False,
            'use_batch_norm': False,
        },
    }
    config = OmegaConf.create(config)

    with pytest.raises(AttributeError):
        tst_rnn_model(config)


def test_all():
    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1) + 1
    x = PaddedBatch(x, length)

    for concat_lens in [False, True]:
        for trainable_starter in ['empty', 'static']:
            for rnn_type in ['gru', 'lstm']:
                for pred_all_states in [False, True]:
                    for pred_all_states_mean in [False, True]:
                        #
                        if pred_all_states and concat_lens:
                            continue
                        if pred_all_states and not pred_all_states_mean:
                            expected_size = 2
                        else:
                            expected_size = 1

                        config = {
                            'rnn': {
                                'trainable_starter': trainable_starter,
                                'hidden_size': 16,
                                'type': rnn_type,
                                'bidir': False,
                            },
                            'head': {
                                'explicit_lengths': concat_lens,
                                'pred_all_states': pred_all_states,
                                'pred_all_states_mean': pred_all_states_mean,
                                'norm_input': False,
                                'use_batch_norm': False,
                            },
                        }
                        config = OmegaConf.create(config)
                        rnn = tst_rnn_model(config)
                        try:
                            out = rnn(x)

                            if isinstance(out, PaddedBatch):
                                out = out.payload

                            assert len(out.size()) == expected_size
                            assert out.size()[0] == 12
                        except Exception:
                            print(config)
                            raise


def test_trx_avg_encoder():
    m = tst_trx_avg_model()

    x = torch.rand(12, 100, 8)
    length = torch.arange(0, 12, 1)
    x = PaddedBatch(x, length)
    out = m(x)

    assert len(out.size()) == 1
    assert out.size()[0] == 12


def test_rnn_iterative_no_starter():
    """
    x_a - clients with 2 parts of transactions: x_a_1 and x_a_2
    x_b_2 - clients with only one second part or transactions

    We run inference in 2 steps:
    x_a_1          -> out_1
    x_a_1 + x_b_2  -> out_2

    Also we run an other kind os splits, for `a` and `b` groups
    x_a   -> out_a
    x_b_2 -> out_b

    We expect the same results in specific positions of out_1, out_2 and out_a, out_b

    :return:
    """
    GRP_A_CLIENT_COUNT = 3
    GRP_B_CLIENT_COUNT = 6
    PART_1_TRX_COUNT = 7
    PART_2_TRX_COUNT = 2
    INPUT_SIZE = 4
    OUTPUT_SIZE = 5

    conf = {
        'hidden_size': OUTPUT_SIZE,
        'type': 'gru',
        'bidir': False,
        'trainable_starter': 'none',
    }
    conf = OmegaConf.create(conf)

    m = RnnEncoder(INPUT_SIZE, **conf)
    m.eval()
    print(m)

    x_a_1 = torch.rand(GRP_A_CLIENT_COUNT, PART_1_TRX_COUNT, INPUT_SIZE)
    x_a_2 = torch.rand(GRP_A_CLIENT_COUNT, PART_2_TRX_COUNT, INPUT_SIZE)
    x_b_2 = torch.rand(GRP_B_CLIENT_COUNT, PART_2_TRX_COUNT, INPUT_SIZE)

    out_a = m(PaddedBatch(torch.cat([x_a_1, x_a_2], dim=1), None))
    out_b = m(PaddedBatch(x_b_2, None))
    out_1 = m(PaddedBatch(x_a_1, None))
    starter = torch.cat([out_1.payload[:, -1, :], torch.zeros(GRP_B_CLIENT_COUNT, OUTPUT_SIZE)], dim=0)
    out_2 = m(PaddedBatch(torch.cat([x_a_2, x_b_2], dim=0), None), starter.unsqueeze(0))

    out_a_merged = torch.cat([out_1.payload, out_2.payload[:GRP_A_CLIENT_COUNT]], dim=1)
    assert ((out_a.payload - out_a_merged).abs() < 1e-4).all()

    out_b_merged = out_2.payload[GRP_A_CLIENT_COUNT:]
    assert ((out_b.payload - out_b_merged).abs() < 1e-4).all()


def test_rnn_iterative_with_starter():
    """
    x_a - clients with 2 parts of transactions: x_a_1 and x_a_2
    x_b_2 - clients with only one second part or transactions

    We run inference in 2 steps:
    x_a_1          -> out_1
    x_a_1 + x_b_2  -> out_2

    Also we run an other kind os splits, for `a` and `b` groups
    x_a   -> out_a
    x_b_2 -> out_b

    We expect the same results in specific positions of out_1, out_2 and out_a, out_b

    :return:
    """
    GRP_A_CLIENT_COUNT = 3
    GRP_B_CLIENT_COUNT = 6
    PART_1_TRX_COUNT = 7
    PART_2_TRX_COUNT = 2
    INPUT_SIZE = 4
    OUTPUT_SIZE = 5

    conf = {
        'hidden_size': OUTPUT_SIZE,
        'type': 'gru',
        'bidir': False,
        'trainable_starter': 'static',
    }
    conf = OmegaConf.create(conf)

    m = RnnEncoder(INPUT_SIZE, **conf)
    m.eval()
    print(m)

    x_a_1 = torch.rand(GRP_A_CLIENT_COUNT, PART_1_TRX_COUNT, INPUT_SIZE)
    x_a_2 = torch.rand(GRP_A_CLIENT_COUNT, PART_2_TRX_COUNT, INPUT_SIZE)
    x_b_2 = torch.rand(GRP_B_CLIENT_COUNT, PART_2_TRX_COUNT, INPUT_SIZE)

    out_a = m(PaddedBatch(torch.cat([x_a_1, x_a_2], dim=1), None))
    out_b = m(PaddedBatch(x_b_2, None))
    out_1 = m(PaddedBatch(x_a_1, None))
    starter = torch.cat([out_1.payload[:, -1, :], torch.zeros(GRP_B_CLIENT_COUNT, OUTPUT_SIZE)], dim=0)
    out_2 = m(PaddedBatch(torch.cat([x_a_2, x_b_2], dim=0), None), starter.unsqueeze(0))

    out_a_merged = torch.cat([out_1.payload, out_2.payload[:GRP_A_CLIENT_COUNT]], dim=1)
    assert ((out_a.payload - out_a_merged).abs() < 1e-4).all()

    out_b_merged = out_2.payload[GRP_A_CLIENT_COUNT:]
    assert ((out_b.payload - out_b_merged).abs() < 1e-4).all()



import torch
import warnings

from ptls.data_load.padded_batch import PaddedBatch
from ptls.nn.trx_encoder.batch_norm import RBatchNorm, RBatchNormWithLens
from ptls.nn.trx_encoder.noisy_embedding import NoisyEmbedding
from ptls.nn.trx_encoder.trx_encoder_base import TrxEncoderBase


class TrxEncoder(TrxEncoderBase):
    """Network layer which makes representation for single transactions

     Input is `PaddedBatch` with ptls-format dictionary, with feature arrays of shape (B, T)
     Output is `PaddedBatch` with transaction embeddings of shape (B, T, H)
     where:
        B - batch size, sequence count in batch
        T - sequence length
        H - hidden size, representation dimension

    `ptls.nn.trx_encoder.noisy_embedding.NoisyEmbedding` implementation are used for categorical features.

    Parameters
        embeddings:
            dict with categorical feature names.
            Values must be like this `{'in': dictionary_size, 'out': embedding_size}`
            These features will be encoded with lookup embedding table of shape (dictionary_size, embedding_size)
            Values can be a `torch.nn.Embedding` implementation
        numeric_values:
            dict with numerical feature names.
            Values must be a string with scaler_name.
            Possible values are: 'identity', 'sigmoid', 'log', 'year'.
            These features will be scaled with selected scaler.
            Values can be `ptls.nn.trx_encoder.scalers.BaseScaler` implementatoin

            One field can have many scalers. In this case key become alias and col name should be in scaler.
            Check `TrxEncoderBase.numeric_values` for more details

        embeddings_noise (float):
            Noise level for embedding. `0` meens without noise
        emb_dropout (float):
            Probability of an element of embedding to be zeroed
        spatial_dropout (bool):
            Whether to dropout full dimension of embedding in the whole sequence

        use_batch_norm:
            True - All numerical values will be normalized after scaling
            False - No normalizing for numerical values
        use_batch_norm_with_lens:
            True - Respect seq_lens during batch_norm. Padding zeroes will be ignored
            False - Batch norm ower all time axis. Padding zeroes will included.

        orthogonal_init:
            if True then `torch.nn.init.orthogonal` applied
        linear_projection_size:
            Linear layer at the end will be added for non-zero value

        out_of_index:
            How to process a categorical indexes which are greater than dictionary size.
            'clip' - values will be collapsed to maximum index. This works well for frequency encoded categories.
                We join infrequent categories to one.
            'assert' - raise an error of invalid index appear.

        norm_embeddings: keep default value for this parameter
        clip_replace_value: Not useed. keep default value for this parameter
        positions: Not used. Keep default value for this parameter

    Examples:
        >>> B, T = 5, 20
        >>> trx_encoder = TrxEncoder(
        >>>     embeddings={'mcc_code': {'in': 100, 'out': 5}},
        >>>     numeric_values={'amount': 'log'},
        >>> )
        >>> x = PaddedBatch(
        >>>     payload={
        >>>         'mcc_code': torch.randint(0, 99, (B, T)),
        >>>         'amount': torch.randn(B, T),
        >>>     },
        >>>     length=torch.randint(10, 20, (B,)),
        >>> )
        >>> z = trx_encoder(x)
        >>> assert z.payload.shape == (5, 20, 6)  # B, T, H
    """
    def __init__(self,
                 embeddings=None,
                 numeric_values=None,
                 embeddings_noise: float = 0,
                 norm_embeddings=None,
                 use_batch_norm=True,
                 use_batch_norm_with_lens=False,
                 clip_replace_value=None,
                 positions=None,
                 emb_dropout=0,
                 spatial_dropout=False,
                 orthogonal_init=False,
                 linear_projection_size=0,
                 out_of_index: str = 'clip',
                 ):
        if clip_replace_value is not None:
            warnings.warn('`clip_replace_value` attribute is deprecated. Always "clip to max" used. '
                          'Use `out_of_index="assert"` to avoid categorical values clip', DeprecationWarning)

        if positions is not None:
            warnings.warn('`positions` is deprecated. positions is not used', UserWarning)

        if embeddings is None:
            embeddings = {}
        if numeric_values is None:
            numeric_values = {}

        noisy_embeddings = {}
        for emb_name, emb_props in embeddings.items():
            if emb_props.get('disabled', False):
                continue
            if emb_props['in'] == 0 or emb_props['out'] == 0:
                continue
            noisy_embeddings[emb_name] = NoisyEmbedding(
                num_embeddings=emb_props['in'],
                embedding_dim=emb_props['out'],
                padding_idx=0,
                max_norm=1 if norm_embeddings else None,
                noise_scale=embeddings_noise,
                dropout=emb_dropout,
                spatial_dropout=spatial_dropout,
            )

        super().__init__(
            embeddings=noisy_embeddings,
            numeric_values=numeric_values,
            out_of_index=out_of_index,
        )

        numerical_size = self.numerical_size
        if use_batch_norm and numerical_size > 0:
            if use_batch_norm_with_lens:
                self.numerical_batch_norm = RBatchNormWithLens(numerical_size)
            else:
                self.numerical_batch_norm = RBatchNorm(numerical_size)
        else:
            self.numerical_batch_norm = None

        if linear_projection_size > 0:
            self.linear_projection_head = torch.nn.Linear(super().output_size, linear_projection_size)
        else:
            self.linear_projection_head = None

        if orthogonal_init:
            for n, p in self.named_parameters():
                if n.startswith('embeddings.') and n.endswith('.weight'):
                    torch.nn.init.orthogonal_(p.data[1:])
                if n == 'linear_projection_head.weight':
                    torch.nn.init.orthogonal_(p.data)

    def forward(self, x: PaddedBatch):
        processed_embeddings = []
        processed_numerical = []

        for field_name in self.embeddings.keys():
            processed_embeddings.append(self.get_category_embeddings(x, field_name))

        for field_name in self.numeric_values.keys():
            processed_numerical.append(self.get_numeric_scaled(x, field_name))

        if len(processed_numerical):
            processed_numerical = torch.cat(processed_numerical, dim=2)
            if self.numerical_batch_norm is not None:
                processed_numerical = PaddedBatch(processed_numerical, x.seq_lens)
                processed_numerical = self.numerical_batch_norm(processed_numerical)
                processed_numerical = processed_numerical.payload
            processed_embeddings.append(processed_numerical)

        out = torch.cat(processed_embeddings, dim=2)

        if self.linear_projection_head is not None:
            out = self.linear_projection_head(out)
        return PaddedBatch(out, x.seq_lens)

    @property
    def output_size(self):
        """Returns hidden size of output representation
        """
        if self.linear_projection_head is not None:
            return self.linear_projection_head.out_features
        return super().output_size



import numpy as np
import torch
import random

from ptls.data_load.utils import collate_feature_dict
from ptls.data_load.feature_dict import FeatureDict

class MlmIndexedDataset(torch.utils.data.Dataset):
    """

    Parameters
    ----------
    data:
        List with all sequences
    seq_len:
        Length of sampled sequence
    step_rate:
        Define step of window moving. `step = seq_len * step_rate`.
        When `step_rate == 1.0` then `step = seq_len` ans windows aren't intersect
        When `step_rate < 1.0` then sampled windows are intersect
        When `step_rate > 1.0` then sampled windows aren't intersect and there are missing transactions
    random_shift:
        Move window start position in (-random_shift, random_shift) interval randomly
    random_crop:
        Reduce lenght of sampled sequence in (0, random_crop) interval randomly
    """
    def __init__(self,
                 data,
                 seq_len: int,
                 step_rate: float = 1.0,
                 random_shift: int = 0,
                 random_crop: int = 0,
                 ):
        self.data = data
        self.seq_len = seq_len
        self.step = int(seq_len * step_rate)
        self.random_shift = random_shift
        self.random_crop = random_crop

        assert self.random_crop < self.seq_len

        self.ix = []
        for item_id, v in enumerate(self.data):
            et = next(iter(v.values()))
            for start_pos in range(0, len(et), self.step):
                self.ix.append([item_id, start_pos])
        self.ix = np.array(self.ix)

    def __len__(self):
        return self.ix.shape[0]

    def __getitem__(self, item):
        item_id, start_pos = self.ix[item]
        v = self.data[item_id]
        seq_len = FeatureDict.get_seq_len(v)

        if self.random_shift > 0:
            start_pos = start_pos + random.randint(-self.random_shift, self.random_shift)
        start_pos = max(start_pos, 0)
        start_pos = min(start_pos, seq_len - self.step)
        len_reduce = 0 if self.random_crop == 0 else random.randint(0, self.random_crop)

        return {k: v[start_pos: start_pos + self.seq_len - len_reduce]
                for k, v in v.items()
                if FeatureDict.is_seq_feature(k, v)}

    @staticmethod
    def collate_fn(batch):
        return collate_feature_dict(batch)

class MLMNSPIndexedDataset(MlmIndexedDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def collate_fn(batch):
        max_lenght = max([len(next(iter(rec.values()))) for rec in batch])

        lefts, rights = [], []
        for rec in batch:
            left, right = sequence_pair_augmentation(rec, max_lenght=max_lenght)
            lefts.append(left)
            rights.append(right)
        
        #
        lefts = lefts * 2
        rights_ = rights[:]
        random.shuffle(rights_)
        rights += rights_

        targets = torch.cat([
            torch.ones(len(batch), dtype=torch.int64),
            torch.zeros(len(batch), dtype=torch.int64),
        ])
        
        concated = [{k: torch.cat([l[k], r[k]]) for k in l.keys()} for l, r in zip(lefts, rights)]
        
        augmented_batch =  padded_collate_wo_target(concated)
        return augmented_batch, targets.float()



"""
Creates data loaders for Sequence Order Prediction learning
Each record pass throw the sampler and splits into views of the main record
Each sample slits into left and right part.
X - is pair of feature_arrays (left and right), sub-samples of main event sequences
Y - order of left and right part. 0 is (left, right), 1 is (right, left)

    Input:        -> Output
dict of arrays    ->  X,                                            Y
client_1          -> (client_1_smpl_1_left, client_1_smpl_1_right), 0
                     (client_1_smpl_2_left, client_1_smpl_2_right), 0
                     (client_1_smpl_3_right, client_1_smpl_3_left), 1
client_2          -> (client_2_smpl_1_right, client_2_smpl_1_left), 1
                     ...
"""
import logging
import warnings

import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from ptls.data_load.augmentations.build_augmentations import build_augmentations
from ptls.data_load.augmentations.sequence_pair_augmentation import sequence_pair_augmentation
from ptls.data_load import IterableChain, augmentation_chain, padded_collate_wo_target
from ptls.data_load.iterable_processing.category_size_clip import CategorySizeClip
from ptls.data_load.iterable_processing.feature_filter import FeatureFilter
from ptls.data_load.iterable_processing.iterable_shuffle import IterableShuffle
from ptls.data_load.iterable_processing.seq_len_filter import SeqLenFilter
from ptls.data_load.list_splitter import ListSplitter
from ptls.data_load.datasets.parquet_dataset import ParquetDataset, ParquetFiles
from ptls.data_load.partitioned_dataset import PartitionedDataset, PartitionedDataFiles
from ptls.metric_learn.dataset import nested_list_to_flat_with_collate
from ptls.frames.coles import split_strategy
from ptls.metric_learn.dataset.splitting_dataset import IterableSplittingDataset, MapSplittingDataset

logger = logging.getLogger(__name__)


def collate_sop_pairs(batch):
    targets = torch.randint(low=0, high=2, size=(len(batch),))

    lefts = [left if target else right for (left, right), target in zip(batch, targets)]
    rights = [right if target else left for (left, right), target in zip(batch, targets)]

    return (
        (
            padded_collate_wo_target(lefts),
            padded_collate_wo_target(rights)
        ),
        targets.long()
    )


class SopDataModuleTrain(pl.LightningDataModule):
    def __init__(self, type, setup, train, valid, pl_module):
        warnings.warn('Use `ptls.frames.PtlsDataModule` '
                      'with `ptls.frames.bert.SopDataset` or `ptls.frames.bert.SopIterableDataset`',
                      DeprecationWarning)
        super().__init__()

        self._type = type
        assert self._type in ('map', 'iterable')

        self.setup_conf = setup
        self.train_conf = train
        self.valid_conf = valid

        self.col_id = self.setup_conf.col_id
        self.category_names = pl_module.seq_encoder.category_names
        self.category_names.add('event_time')
        self.category_max_size = pl_module.seq_encoder.category_max_size

        self.train_dataset = None
        self.valid_dataset = None
        self._train_ids = None
        self._valid_ids = None

    def prepare_data(self):
        if 'dataset_files' in self.setup_conf:
            self.setup_iterable_files()
        elif 'dataset_parts' in self.setup_conf:
            self.setup_iterable_parts()
        else:
            raise AttributeError(f'Missing dataset definition. One of `dataset_parts` or `dataset_files` expected')

        if self._type == 'map':
            self.setup_map()

    def setup_iterable_files(self):
        if self.setup_conf.split_by == 'files':
            data_files = ParquetFiles(self.setup_conf.dataset_files.data_path).data_files

            splitter = ListSplitter(
                data_files,
                valid_size=self.setup_conf.valid_size,
                seed=self.setup_conf.valid_split_seed,
            )
            logger.info(f'Prepared splits: '
                        f'{len(splitter.train)} files in train, {len(splitter.valid)} files in valid')

            self.train_dataset = ParquetDataset(
                splitter.train,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = ParquetDataset(
                splitter.valid,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )

        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def setup_iterable_parts(self):
        data_files = PartitionedDataFiles(**self.setup_conf.dataset_parts)

        if self.setup_conf.split_by == 'hash_id':
            splitter = ListSplitter(
                data_files.hash_parts,
                valid_size=self.setup_conf.valid_size,
                seed=self.setup_conf.valid_split_seed,
            )
            logger.info(f'Prepared splits: '
                        f'{len(data_files.dt_parts)} parts in dt_train, {len(data_files.dt_parts)} parts in dt_valid, '
                        f'{len(splitter.train)} parts in hash_train, {len(splitter.valid)} parts in hash_valid')

            self.train_dataset = PartitionedDataset(
                data_files.data_path, data_files.dt_parts, splitter.train, col_id=self.col_id,
                post_processing=IterableChain(*self.build_iterable_processing('train')),
                shuffle_files=True if self._type == 'iterable' else False,
            )
            self.valid_dataset = PartitionedDataset(
                data_files.data_path, data_files.dt_parts, splitter.valid, col_id=self.col_id,
                post_processing=IterableChain(*self.build_iterable_processing('valid')),
                shuffle_files=False,
            )
        elif self.setup_conf.split_by == 'rows':
            raise NotImplementedError("Split by rows aren't supported for partitioned dataset")
        else:
            raise AttributeError(f'Unknown split strategy: {self.setup_conf.split_by}')

    def build_iterable_processing(self, part):
        if part == 'train':
            yield SeqLenFilter(min_seq_len=self.train_conf.min_seq_len)

        yield FeatureFilter(keep_feature_names=self.category_names)
        yield CategorySizeClip(self.category_max_size)

        if self._type == 'iterable':
            # all processing in single chain
            if part == 'train':
                yield IterableShuffle(self.train_conf.buffer_size)

            if part == 'train':
                split_strategy_conf = self.train_conf.split_strategy
            else:
                split_strategy_conf = self.valid_conf.split_strategy
            yield IterableSplittingDataset(split_strategy.create(**split_strategy_conf),
                                           self.build_augmentations(part))

    def build_augmentations(self, part):
        if part == 'train':
            a_chain = build_augmentations(self.train_conf.augmentations)
        else:
            a_chain = build_augmentations(self.valid_conf.augmentations)
        return augmentation_chain(a_chain, sequence_pair_augmentation)

    def setup_map(self):
        self.train_dataset = list(tqdm(iter(self.train_dataset)))
        logger.info(f'Loaded {len(self.train_dataset)} for train')
        self.valid_dataset = list(tqdm(iter(self.valid_dataset)))
        logger.info(f'Loaded {len(self.valid_dataset)} for valid')

        self.train_dataset = MapSplittingDataset(
            base_dataset=self.train_dataset,
            splitter=split_strategy.create(**self.train_conf.split_strategy),
            a_chain=self.build_augmentations('train'),
        )
        self.valid_dataset = MapSplittingDataset(
            base_dataset=self.valid_dataset,
            splitter=split_strategy.create(**self.valid_conf.split_strategy),
            a_chain=self.build_augmentations('valid'),
        )

    def train_dataloader(self):
        return DataLoader(
            dataset=self.train_dataset,
            collate_fn=nested_list_to_flat_with_collate(collate_sop_pairs),
            shuffle=False if self._type == 'iterable' else True,
            num_workers=self.train_conf.num_workers,
            batch_size=self.train_conf.batch_size,
        )

    def val_dataloader(self):
        return DataLoader(
            dataset=self.valid_dataset,
            collate_fn=nested_list_to_flat_with_collate(collate_sop_pairs),
            num_workers=self.valid_conf.num_workers,
            batch_size=self.valid_conf.batch_size,
        )



from functools import reduce
from operator import iadd

import torch

from ptls.data_load.augmentations.sequence_pair_augmentation import sequence_pair_augmentation
from ptls.data_load.utils import collate_feature_dict
from ptls.frames.coles import ColesDataset


class SopDataset(ColesDataset):
    def get_splits(self, feature_arrays):
        return [sequence_pair_augmentation(item)
                for item in super().get_splits(feature_arrays)]

    @staticmethod
    def collate_fn(batch):
        batch = reduce(iadd, batch)
        targets = torch.randint(low=0, high=2, size=(len(batch),))

        lefts = [left if target else right for (left, right), target in zip(batch, targets)]
        rights = [right if target else left for (left, right), target in zip(batch, targets)]

        return (
            (
                collate_feature_dict(lefts),
                collate_feature_dict(rights)
            ),
            targets.long()
        )


class SopIterableDataset(SopDataset, torch.utils.data.IterableDataset):
    pass



import numpy as np
import torch

from ptls.data_load import IterableProcessingDataset


class TakeFirstTrx(IterableProcessingDataset):
    def __init__(self, take_first_fraction=0.5, seq_len_col=None, sequence_col=None):
        """
        Args:
            take_first_fraction: control the fraction of transactions to keep
                                 EXAMPLE: take_first_fraction=0.75 -> the last 0.25 of all user
                                          transactions will be chosen as user target distribution
                                          and therefore will be cutted off
            seq_len_col:  field where sequence length stored, if None, `target_col` used
            sequence_col: field for sequence length detection, if None, any
                          iterable field will be used
        """
        super().__init__()

        self._take_first_fraction = take_first_fraction
        self._sequence_col = sequence_col
        self._seq_len_col = seq_len_col

    def __iter__(self):
        for rec in self._src:
            features = rec[0] if type(rec) is tuple else rec
            seq_len = self.get_len(features)
            take_first_n = int(seq_len * self._take_first_fraction)
            for key, val in features.items():
                if self.is_seq_feature(key, val):
                    features[key] = val[:take_first_n]
            rec = (features, rec[1]) if type(rec) is tuple else features
            yield rec

    def get_sequence_col(self, rec):
        if self._sequence_col is None:
            arrays = [k for k, v in rec.items() if self.is_seq_feature(k, v)]
            if len(arrays) == 0:
                raise ValueError(f'Can not find field with sequence from record: {rec}')
            self._sequence_col = arrays[0]
        return self._sequence_col

    def get_len(self, rec):
        if self._seq_len_col is not None:
            return rec[self._seq_len_col]
        return len(rec[self.get_sequence_col(rec)])



from typing import Dict

import numpy as np
import torch

from ptls.data_load.feature_dict import FeatureDict


class PaddedBatch:
    """Contains a padded batch of sequences with different lengths.

    Parameters:
        payload:
            container with data. Format supported:
            - dict with features. This is the input data for overall network pipeline.
                Kees are the feature names, values are (B, T) shape tensors.
                Long type for categorical features, embedding lookup table indexes expected
                Float type for numerical features.
            - trx embedding tensor. This is the intermediate data for overall network pipeline.
                shape (B, T, H)
            - feature tensor. Used in some cases
                shape (B, T)
        length:
            Tensor of shape (B,) with lengths of sequences.
            All sequences in `payload` has length T, but only L first are used.
            Unused positions padded with zeros

    Example:
        >>> data = PaddedBatch(
        >>>     payload=torch.tensor([
        >>>         [1, 2, 0, 0],
        >>>         [3, 4, 5, 6],
        >>>         [7, 8, 9, 0],
        >>>     ]),
        >>>     length=torch.tensor([2, 4, 3]),
        >>> )
        >>>
        >>> # check shape
        >>> torch.testing.assert_close(data.payload.size(), (3, 4))
        >>>
        >>> # get first transaction
        >>> torch.testing.assert_close(data.payload[:, 0], torch.tensor([1, 3, 7]))
        >>>
        >>> # get last transaction
        >>> torch.testing.assert_close(data.payload[torch.arange(3), data.seq_lens - 1], torch.tensor([2, 6, 9]))
        >>>
        >>> # get all transaction flatten
        >>> torch.testing.assert_close(data.payload[data.seq_len_mask.bool()], torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9]))

    """
    def __init__(self, payload: Dict[str, torch.Tensor], length: torch.LongTensor):
        self._payload = payload
        self._length = length

    @property
    def payload(self):
        return self._payload

    @property
    def seq_lens(self):
        return self._length

    @property
    def device(self):
        return self._length.device

    @property
    def seq_feature_shape(self):
        return next(v.size() for k, v in self._payload.items() if self.is_seq_feature(k, v))

    def __len__(self):
        return len(self._length)

    def to(self, device, non_blocking=False):
        length = self._length.to(device=device, non_blocking=non_blocking)
        payload = {
            k: v.to(device=device, non_blocking=non_blocking) if type(v) is torch.Tensor else v
            for k, v in self._payload.items()
        }
        return PaddedBatch(payload, length)

    @property
    def seq_len_mask(self):
        """mask with B*T size for valid tokens in `payload`
        """
        if type(self._payload) is dict:
            B, T = next(v for k, v in self._payload.items() if self.is_seq_feature(k, v)).size()
        else:
            B, T = self._payload.size()[:2]
        return (torch.arange(T, device=self._length.device).unsqueeze(0).expand(B, T) < \
                self._length.unsqueeze(1)).long()

    @staticmethod
    def is_seq_feature(k: str, x):
        """Check is value sequential feature
        Synchronized with ptls.data_load.feature_dict.FeatureDict.is_seq_feature

                     1-d        2-d
        event_time | True      True
        target_    | False     False  # from FeatureDict.is_seq_feature
        tensor     | False     True

        Parameters
        ----------
        k:
            feature_name
        x:
            value for check

        Returns
        -------

        """
        if k == 'event_time':
            return True
        if k.startswith('target'):
            return False
        if type(x) is np.ndarray:
                return False
        if type(x) is torch.Tensor and len(x.shape) == 1:
            return False
        return True 

    def drop_seq_features(self):
        """Returns new dict without sequential features

        Returns
        -------

        """
        return {k: v for k, v in self.payload.items() if not PaddedBatch.is_seq_feature(k, v)}

    def keep_seq_features(self):
        """Returns new PaddedBatch with sequential features only

        Returns
        -------

        """
        return PaddedBatch(
            payload={k: v for k, v in self.payload.items() if PaddedBatch.is_seq_feature(k, v)},
            length=self.seq_lens,
        )



import hydra
import torch
from omegaconf import OmegaConf
import pandas as pd
from sklearn.metrics import roc_auc_score
import numpy as np

from ptls.data_load import PaddedBatch
from ptls.data_load.utils import collate_feature_dict
from ptls.frames.inference_module import InferenceModule
import pytorch_lightning as pl

from ptls.nn import PBFeatureExtract
from ptls_tests.utils.data_generation import gen_trx_data


def get_rnn_seq_encoder():
    conf = OmegaConf.create("""
_target_: torch.nn.Sequential
_args_:
  - _target_: ptls.nn.RnnSeqEncoder
    trx_encoder:
      _target_: ptls.nn.TrxEncoder
      embeddings:
        mcc_code:
          in: 21
          out: 3
        trans_type:
          in: 11
          out: 2
      numeric_values: 
        amount: log
    hidden_size: 16
  - _target_: ptls.nn.Head
    input_size: 16
    objective: classification
    num_classes: 1
    """)
    return hydra.utils.instantiate(conf)


def test_inference_module_predict():
    valid_loader = torch.utils.data.DataLoader(
        gen_trx_data((torch.rand(1000)*60+1).long(), target_type='bin_cls', use_feature_arrays_key=False),
        batch_size=64, collate_fn=collate_feature_dict)
    rnn_model = InferenceModule(
        model=get_rnn_seq_encoder(),
        model_out_name='pred',
    )

    df_out = pd.concat(pl.Trainer(gpus=0, max_epochs=-1).predict(rnn_model, valid_loader))
    print(roc_auc_score(df_out['target'], df_out['pred']))


def test_score_model_mult2():
    valid_loader = [
        PaddedBatch(
            {'embeddings': torch.rand(4, 16), 'target_int': torch.arange(4), 'target_str': np.arange(4).astype(str)},
            torch.LongTensor([1, 3, 16, 5]),
        ),
        PaddedBatch(
            {'embeddings': torch.rand(2, 16), 'target_int': torch.arange(2), 'target_str': np.arange(2).astype(str)},
            torch.LongTensor([16, 2]),
        ),
        PaddedBatch(
            {'embeddings': torch.rand(1, 16), 'target_int': torch.arange(1), 'target_str': np.arange(1).astype(str)},
            torch.LongTensor([16]),
        ),
    ]
    model = InferenceModule(
        model=torch.nn.Sequential(
            PBFeatureExtract('embeddings', as_padded_batch=False),
            torch.nn.Linear(16, 2),
            torch.nn.Sigmoid(),
        ),
        model_out_name='pred', pandas_output=False,
    )

    dict_out = pl.Trainer(gpus=0, max_epochs=-1).predict(model, iter(valid_loader))
    id1 = torch.cat([v['target_int'] for v in dict_out])
    id2 = np.concatenate([v['target_str'] for v in dict_out])

    assert torch.cat([v['pred'] for v in dict_out], dim=0).shape == (7, 2)
    assert id1.shape == (7,)
    assert id2.shape == (7,)

    np.testing.assert_array_almost_equal(id1, np.array([0, 1, 2, 3, 0, 1, 0]))
    assert id2.tolist() == ['0', '1', '2', '3', '0', '1', '0']



import numpy as np


def _get_dtype(max_size):
    if max_size <= 2 ** 8 - 1:
        return np.uint8, max_size
    if max_size <= 2 ** (16 - 1) - 1:
        return np.int16, max_size
    if max_size <= 2 ** (32 - 1) - 1:
        return np.int32, max_size
    return np.int64, max_size


def _dtype_mapping(embeddings, numeric_values):
    dtype_mapping = {k: _get_dtype(v['in']) for k, v in embeddings.items()}
    for k in numeric_values.keys():
        dtype_mapping[k] = np.int32, None
    return dtype_mapping


def _dtype_adjust(k, v, dtype_mapping):
    if k in dtype_mapping:
        dtype, max_value = dtype_mapping[k]
        if max_value is not None:
            v = np.where(
                v < max_value,
                v,
                1
            )
        return v.astype(dtype)

    if v.dtype == np.int8:
        return v.astype(np.int16)
    return v


def feature_local_day(feature_arrays, et):
    feature_arrays['local_day'] = ((et.astype('datetime64[D]') -
                                    et.astype('datetime64[M]')).astype(np.int16) + 1)


def feature_local_month(feature_arrays, et):
    feature_arrays['local_month'] = (et.astype('datetime64[M]').astype(np.int16) % 12 + 1)


def feature_local_weekday(feature_arrays, et):
    feature_arrays['local_weekday'] = ((et.astype('datetime64[D]').astype(np.int16) + 3) % 7 + 1)


def feature_hour(feature_arrays, et):
    feature_arrays['hour'] = ((et.astype('datetime64[s]') -
                               et.astype('datetime64[D]')).astype(np.int32) // 3600 + 1).astype(np.int16)


def feature_random_number(feature_arrays, et):
    feature_arrays['random_number'] = np.random.random(len(et))


def types_and_features(seq, embeddings, numeric_values):
    dtype_mapping = _dtype_mapping(embeddings, numeric_values)

    feature_list = []
    if 'local_day' in dtype_mapping:
        feature_list.append(feature_local_day)
    if 'local_month' in dtype_mapping:
        feature_list.append(feature_local_month)
    if 'local_weekday' in dtype_mapping:
        feature_list.append(feature_local_weekday)
    if 'hour' in dtype_mapping:
        feature_list.append(feature_hour)
    if 'random_number' in dtype_mapping:
        feature_list.append(feature_random_number)

    for rec in seq:
        et = rec['event_time']

        feature_arrays = rec['feature_arrays']

        # time features
        for f in feature_list:
            f(feature_arrays, et)

        # drop unused keys
        for k in set(feature_arrays) - set(dtype_mapping):
            del feature_arrays[k]

        # fill missing keys for empty records
        for k in dtype_mapping:
            if k not in feature_arrays:
                feature_arrays[k] = np.array([])

        # dtype_adjust
        feature_arrays = {k: _dtype_adjust(k, v, dtype_mapping) for k, v in feature_arrays.items()}

        rec['feature_arrays'] = feature_arrays
        yield rec


def fit_features(seq, embeddings, numeric_values):
    dtype_mapping = _dtype_mapping(embeddings, numeric_values)

    feature_list = []
    if 'local_day' in dtype_mapping:
        feature_list.append(feature_local_day)
    if 'local_month' in dtype_mapping:
        feature_list.append(feature_local_month)
    if 'local_weekday' in dtype_mapping:
        feature_list.append(feature_local_weekday)
    if 'hour' in dtype_mapping:
        feature_list.append(feature_hour)
    if 'random_number' in dtype_mapping:
        feature_list.append(feature_random_number)

    for rec in seq:
        et = rec['event_time']

        feature_arrays = rec['feature_arrays']

        # time features
        for f in feature_list:
            f(feature_arrays, et)

        # drop unused keys
        for k in set(feature_arrays) - set(dtype_mapping):
            del feature_arrays[k]

        # fill missing keys for empty records
        for k in dtype_mapping:
            if k not in feature_arrays:
                feature_arrays[k] = np.array([])

        rec['feature_arrays'] = feature_arrays
        yield rec


def fit_types(seq, embeddings, numeric_values):
    dtype_mapping = _dtype_mapping(embeddings, numeric_values)

    for rec in seq:
        feature_arrays = rec['feature_arrays']

        feature_arrays = {k: _dtype_adjust(k, v, dtype_mapping) for k, v in feature_arrays.items()}

        rec['feature_arrays'] = feature_arrays
        yield rec


def add_ticks(seq, conf, mode):
    assert mode in ('application_date', 'time_window')

    freq = conf.params.tick_params.freq

    if mode == 'application_date':
        application_start_date = np.datetime64(conf.dataset.application_start_date)
        transaction_start_date = np.datetime64(conf.dataset.transaction_start_date)
        max_days = conf.params.max_days
        trx_start_date = max(application_start_date - np.timedelta64(max_days, 'D'), transaction_start_date)
        trx_end_date = np.datetime64(conf.dataset.ripe_date)
    elif mode == 'time_window':
        trx_start_date = np.datetime64(conf.dataset.start_date)
        trx_end_date = np.datetime64(conf.dataset.end_date)

    all_dates = np.arange(trx_start_date, trx_end_date, np.timedelta64(1, 'D'))
    if freq == 'w':
        weekdays = ((all_dates.astype('datetime64[D]').astype(np.int16) + 3) % 7 + 1)
        all_dates = all_dates[weekdays == 1]
    elif freq == 'm':
        monthdays = (all_dates.astype('datetime64[D]') - all_dates.astype('datetime64[M]') + 1).astype('int')
        all_dates = all_dates[monthdays == 15]
    else:
        raise AttributeError(f'Only "w" - week and "m" - month, freq supported. Found "{freq}"')

    all_zeros = np.zeros(len(all_dates), dtype=np.int16)

    for rec in seq:
        if mode == 'application_date':
            application_date = rec['application_date']

        et = rec['event_time']
        original_len = len(et)
        feature_arrays = rec['feature_arrays']

        if mode == 'application_date':
            add_dates = all_dates[all_dates <= application_date]
            add_dates = add_dates[add_dates >= application_date - np.timedelta64(max_days, 'D')]
            add_features = all_zeros[:len(add_dates)]
        elif mode == 'time_window':
            add_dates = all_dates
            add_features = all_zeros

        if len(et) > 0:
            et = np.concatenate([et, add_dates])
        else:
            et = add_dates
        ix_to_take = np.argsort(et)

        if 'tick' in feature_arrays:
            del feature_arrays['tick']
        feature_arrays = {k: np.concatenate([v, add_features]) for k, v in feature_arrays.items()}
        et = et[ix_to_take]
        feature_arrays = {k: v[ix_to_take] for k, v in feature_arrays.items()}
        feature_arrays['tick'] = (ix_to_take >= original_len).astype(np.int16)

        rec['feature_arrays'] = feature_arrays
        rec['event_time'] = et

        yield rec
